# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import numpy as np
from pydantic import BaseModel, Field, model_validator

from tesseract_core.runtime import Array, Differentiable, Float32, Int32, ShapeDType

#
# Schemas
#


class VolumetricMeshData(BaseModel):
    """Mock mesh schema; shapes not validated."""

    n_points: int
    n_cells: int

    points: Differentiable[Array[(None, 3), Float32]]
    num_points_per_cell: Array[(None,), Float32]  # should have length == n_cells
    cell_connectivity: Array[(None,), Int32]  # length == sum(num_points_per_cell)

    cell_data: dict[str, Array[(None, None), Float32]]
    point_data: dict[str, Array[(None, None), Float32]]

    @model_validator(mode="after")
    def validate_num_points_per_cell(self):
        if not isinstance(self.num_points_per_cell, np.ndarray):
            return self
        if len(self.num_points_per_cell) != self.n_cells:
            raise ValueError(f"Length of num_points_per_cell must be {self.n_cells}")
        return self

    @model_validator(mode="after")
    def validate_cell_connectivity(self):
        if not isinstance(self.cell_connectivity, np.ndarray):
            return self
        expected_len = sum(self.num_points_per_cell)
        if len(self.cell_connectivity) != expected_len:
            raise ValueError(f"Length of cell_connectivity must be {expected_len}")
        return self


class InputSchema(BaseModel):
    mesh: VolumetricMeshData = Field(
        description="The mesh you want summary statistics of"
    )


class SummaryStatistics(BaseModel):
    first_point_coordinates: Differentiable[Array[(3,), Float32]] = Field(
        description="Coordinates of the first point defined in the mesh."
    )
    barycenter: Differentiable[Array[(3,), Float32]] = Field(
        "Mean of all the points defined in the input mesh."
    )


class OutputSchema(BaseModel):
    statistics: SummaryStatistics = Field(description="Summary statistics of the mesh.")


#
# Required endpoints
#


def apply(inputs: InputSchema) -> OutputSchema:
    points = inputs.mesh.points

    statistics = SummaryStatistics(
        first_point_coordinates=points[0],
        barycenter=points.mean(axis=0),
    )

    return OutputSchema(statistics=statistics)


#
# Optional endpoints
#


def abstract_eval(abstract_inputs):
    input_points_shapedtype = abstract_inputs.mesh.points

    # get dimension of vector space points live in from input
    dim = input_points_shapedtype.shape[1]
    dtype = input_points_shapedtype.dtype
    return {
        "statistics": {
            "first_point_coordinates": ShapeDType(shape=(dim,), dtype=dtype),
            "barycenter": ShapeDType(shape=(dim,), dtype="float32"),
        }
    }


def jacobian(inputs: InputSchema, jac_inputs: set[str], jac_outputs: set[str]):
    # TODO: Make this work with other inputs as well
    assert tuple(jac_inputs) == ("mesh.points",)

    jac = {}

    if "statistics.barycenter" in jac_outputs:
        jac_dbar_dpoints = np.zeros((3, inputs.mesh.n_points, 3))
        jac_dbar_dpoints[0, :, 0] = 1.0 / inputs.mesh.n_points
        jac_dbar_dpoints[1, :, 1] = 1.0 / inputs.mesh.n_points
        jac_dbar_dpoints[2, :, 2] = 1.0 / inputs.mesh.n_points
        jac["statistics.barycenter"] = {"mesh.points": jac_dbar_dpoints}

    if "statistics.first_point_coordinates" in jac_outputs:
        jac_dfpc_dpoints = np.zeros((3, inputs.mesh.n_points, 3))
        jac_dfpc_dpoints[0, 0, 0] = 1.0
        jac_dfpc_dpoints[1, 0, 1] = 1.0
        jac_dfpc_dpoints[2, 0, 2] = 1.0

        jac["statistics.first_point_coordinates"] = {"mesh.points": jac_dfpc_dpoints}

    return jac



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Iterator, Sequence
from pathlib import Path
from typing import (
    Annotated,
    Any,
    Callable,
    Union,
    get_args,
    get_origin,
)

from pydantic import (
    AfterValidator,
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
    TypeAdapter,
)
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import SchemaSerializer, SchemaValidator, core_schema

from tesseract_core.runtime.file_interactions import parent_path
from tesseract_core.runtime.mpa import (
    log_artifact,
    log_metric,
    log_parameter,
    start_run,
)
from tesseract_core.runtime.schema_types import safe_issubclass


class LazySequence(Sequence):
    """Lazy sequence type that loads items from a file handle on access.

    This allows users to define a sequence of objects that are lazily loaded from a data source,
    and validated when accessed.

    When used as a Pydantic annotation, lazy sequences accept either a list of objects or a
    glob pattern to load objects from a file path.

    Example:
        >>> class MyModel(BaseModel):
        ...     objects: LazySequence[str]
        >>> model = MyModel.model_validate({"objects": ["item1", "item2"]})
        >>> model.objects[0]
        'item1'
        >>> model = MyModel.model_validate({"objects": "@/path/to/data/*.json"})
        >>> model.objects[1]
        'item2'
    """

    def __init__(self, keys: Sequence[Any], getter: Callable[[Any], Any]) -> None:
        """Initialize a LazySequence with the given keys and getter function.

        Args:
            keys: Sequence of keys to load items from.
            getter: Function that loads an item from a key.

        Example:
            >>> items = LazySequence(["item1", "item2"], lambda key: f"Loaded {key}")
            >>> items[0]
            'Loaded item1'
        """
        self.keys = keys
        self.getter = getter

    def __class_getitem__(cls, base_type: type) -> type:
        """Create a new type annotation based on the given wrapped type."""
        # Support for LazySequence[MyObject] syntax
        return Annotated[Sequence[base_type], PydanticLazySequenceAnnotation]

    @classmethod
    def __get_pydantic_core_schema__(cls, *args: Any, **kwargs: Any) -> None:
        # Raise if LazySequence is accidentally used as Pyedantic annotation without a wrapped type
        raise NotImplementedError(
            f"Generic {cls.__name__} objects do not support Pydantic schema generation. "
            f"Did you mean to use {cls.__name__}[MyObject]?"
        )

    def __getitem__(self, key: int) -> Any:
        if not isinstance(key, int):
            raise TypeError("LazySequence indices must be integers")
        return self.getter(self.keys[key])

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(keys={self.keys})"

    def __len__(self) -> int:
        return len(self.keys)

    def __iter__(self) -> Iterator[Any]:
        return (self.__getitem__(idx) for idx in range(len(self)))


class PydanticLazySequenceAnnotation:
    """Pydantic annotation for lazy sequences."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        raise RuntimeError(f"{self.__class__.__name__} cannot be instantiated")

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        """This method is called by Pydantic to get the core schema for the annotated type.

        Does most of the heavy lifting for validation and serialization.
        """

        def create_sequence(maybe_path: Union[str, Sequence[Any]]) -> LazySequence:
            """Expand a glob pattern into a LazySequence if needed."""
            validator = SchemaValidator(item_schema)

            if not isinstance(maybe_path, str) or not maybe_path.startswith("@"):
                items = maybe_path
                getter = validator.validate_python
                return LazySequence(items, getter)

            # We know that the path is a glob pattern, so we need to load items from files
            from .file_interactions import (
                expand_glob,
                guess_format_from_path,
                load_bytes,
                read_from_path,
            )

            maybe_path = maybe_path[1:]
            items = expand_glob(maybe_path)

            def load_item(key: str) -> Any:
                buffer = read_from_path(key)
                format = guess_format_from_path(key)
                obj = load_bytes(buffer, format)
                context = {"base_dir": parent_path(key)}
                return validator.validate_python(obj, context=context)

            return LazySequence(items, load_item)

        def serialize(obj: LazySequence, __info: Any) -> Any:
            """When serializing, convert the LazySequence to a list of items.

            This is not an encouraged use case, but it is supported for completeness.
            """
            materialized_sequence = list(obj)
            serializer = SchemaSerializer(sequence_schema)

            return serializer.to_python(materialized_sequence, **__info.__dict__)

        origin = get_origin(_source_type)
        if not safe_issubclass(origin, Sequence):
            # should never happen, since we always use Annotated[Sequence[...], PydanticLazySequenceAnnotation]
            raise ValueError(
                f"LazySequence can only be used with Sequence types, not {origin}"
            )

        # This is a Sequence, so args is a single type
        args = get_args(_source_type)
        assert len(args) == 1

        # Wrap in TypeAdapter so we don't need conditional logic for Python types vs. Pydantic models
        item_schema = TypeAdapter(args[0]).core_schema
        sequence_schema = _handler(_source_type)

        obj_or_path = core_schema.union_schema(
            [sequence_schema, core_schema.str_schema(pattern="^@")]
        )
        load_schema = core_schema.chain_schema(
            # first load data, then validate it with the wrapped schema
            [
                obj_or_path,
                core_schema.no_info_plain_validator_function(
                    create_sequence,
                    serialization=core_schema.plain_serializer_function_ser_schema(
                        serialize,
                        info_arg=True,
                        return_schema=sequence_schema,
                    ),
                ),
            ]
        )
        return core_schema.json_or_python_schema(
            json_schema=load_schema,
            python_schema=load_schema,
            serialization=core_schema.plain_serializer_function_ser_schema(
                serialize,
                info_arg=True,
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        """This method is called by Pydantic to get the JSON schema for the annotated type."""
        return handler(_core_schema)


def _resolve_input_path(path: Path) -> Path:
    from tesseract_core.runtime.config import get_config

    input_path = get_config().input_path
    tess_path = (input_path / path).resolve()
    if str(input_path) not in str(tess_path):
        raise ValueError(
            f"Invalid input file reference: {path}. "
            "File references have to be relative to --input-path."
        )
    if not tess_path.exists():
        raise FileNotFoundError(f"Input path {tess_path} does not exist.")
    if not tess_path.is_file():
        raise ValueError(f"Input path {tess_path} is not a file.")
    return tess_path


def _strip_output_path(path: Path) -> Path:
    from tesseract_core.runtime.config import get_config

    output_path = get_config().output_path
    if path.is_relative_to(output_path):
        return path.relative_to(output_path)
    else:
        return path


InputFileReference = Annotated[Path, AfterValidator(_resolve_input_path)]
OutputFileReference = Annotated[Path, AfterValidator(_strip_output_path)]


__all__ = [
    "InputFileReference",
    "LazySequence",
    "OutputFileReference",
    "PydanticLazySequenceAnnotation",
    "log_artifact",
    "log_metric",
    "log_parameter",
    "start_run",
]



import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LogNorm


def make_animation(xs, ys):
    """Animate optimization trajectory on a countour of the rosenbrock function."""
    # Generate a grid of points for the contour plot
    x = np.linspace(-2, 2, 400)
    y = np.linspace(-0.5, 1.5, 400)
    X, Y = np.meshgrid(x, y)
    Z = (1 - X) ** 2 + 100 * (Y - X**2) ** 2

    # Set up the figure and axis
    fig, ax = plt.subplots(figsize=(8, 6))
    levels = np.logspace(-2.5, 3.5, 10)
    ax.contourf(X, Y, Z, levels=levels, norm=LogNorm(), cmap="viridis")
    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.plot(1, 1, "*", markersize=15, label="Global Minimum", color="black")
    (line,) = ax.plot([], [], ".-", markersize=5, label="Trajectory", color="black")
    ax.legend()
    plt.tight_layout()

    def init():
        line.set_data([], [])
        return (line,)

    def update(frame):
        line.set_data(xs[: frame + 1], ys[: frame + 1])
        return (line,)

    return animation.FuncAnimation(
        fig, update, frames=len(xs), init_func=init, interval=300, blit=True
    )



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import cowsay
from pydantic import BaseModel, Field


class InputSchema(BaseModel):
    message: str = Field(description="A message for cowsay.")


class OutputSchema(BaseModel):
    cowsays: str = Field(description="The cowsay output string.")


def apply(inputs: InputSchema) -> OutputSchema:
    """Greet a person whose name is given as input."""
    return OutputSchema(cowsays=cowsay.get_output_string("cow", inputs.message))



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import re
import shutil
from pathlib import Path

from tesseract_core import __version__

# Set the TESSERACT_API_PATH environment variable to the dummy Tesseract API
# This will be used to instantiate the Tesseract runtime so we can generate api docs
os.environ["TESSERACT_API_PATH"] = os.path.abspath(
    os.path.join(
        os.path.dirname(__file__), "..", "tests", "dummy_tesseract", "tesseract_api.py"
    )
)

project = "Tesseract"
copyright = "2025, Pasteur Labs"
author = "The Tesseract Team @ Pasteur Labs + OSS contributors"

# The short X.Y version
parsed_version = re.match(r"(\d+\.\d+\.\d+)", __version__)
if parsed_version:
    version = parsed_version.group(1)
else:
    version = "0.0.0"

# The full version, including alpha/beta/rc tags
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_nb",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
    "sphinx_click.ext",
    # Copy button for code blocks
    "sphinx_copybutton",
    # OpenGraph metadata for social media sharing
    "sphinxext.opengraph",
    # For tab-set directive
    "sphinx_design",
    # For nice rendering of Pydantic models
    "sphinxcontrib.autodoc_pydantic",
]


myst_enable_extensions = [
    "dollarmath",
    "colon_fence",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("http://docs.scipy.org/doc/numpy/", None),
}

templates_path = []
exclude_patterns = ["build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["static"]
html_theme_options = {
    "light_logo": "logo-light.png",
    "dark_logo": "logo-dark.png",
    "sidebar_hide_name": True,
}
html_css_files = ["custom.css"]


# -- Custom directives ----------------------------------------------------


def zip_examples_folder() -> None:
    """Zip a folder and save it to the specified path."""
    import shutil
    from pathlib import Path

    here = Path(__file__).parent

    root_dir = (here / "..").resolve()
    archive_path = here / "downloads" / "examples.zip"

    shutil.make_archive(archive_path.with_suffix(""), "zip", root_dir, "examples")
    assert archive_path.exists()


def setup(app) -> None:
    """Sphinx setup function. Used to register custom stuff."""
    # HACK: We zip the examples folder here so that it can be downloaded
    zip_examples_folder()


# -- Handle Jupyter notebooks ------------------------------------------------

# Do not execute notebooks during build (just take existing output)
nb_execution_mode = "off"

# Copy example notebooks to auto_examples folder on every build
for example_notebook in Path("../demo").glob("*/demo.ipynb"):
    # Copy the example notebook to the docs folder
    dest = (Path("content/demo") / example_notebook.parent.name).with_suffix(".ipynb")
    shutil.copyfile(example_notebook, dest)



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from helloworld import greet
from pydantic import BaseModel, Field


class InputSchema(BaseModel):
    name: str = Field(description="Name of the person you want to greet.")


class OutputSchema(BaseModel):
    greeting: str = Field(description="A greeting!")


def apply(inputs: InputSchema) -> OutputSchema:
    """Greet a person whose name is given as input."""
    return OutputSchema(greeting=greet(inputs.name))



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from pydantic import BaseModel

from tesseract_core.runtime import Differentiable, Float32

#
# Schemas
#


class InputSchema(BaseModel):
    # Make sure that array types work for older Python versions
    foo: Differentiable[Float32]


class OutputSchema(BaseModel):
    bar: Differentiable[Float32]


#
# Required endpoints
#


def apply(inputs: InputSchema) -> OutputSchema:
    # Ensure that the Python version is what we expect (3.9)
    import sys

    assert sys.version_info[:2] == (3, 9)
    return OutputSchema(bar=0)


#
# Optional endpoints
#


def abstract_eval(abstract_inputs):
    return {"bar": {"shape": (), "dtype": "float32"}}


def jacobian(
    inputs: InputSchema, jac_inputs: list[str], jac_outputs: list[str]
) -> OutputSchema:
    return {out: 0.0 for out in jac_outputs}



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import base64
import json
from pathlib import Path

import jsf
import numpy as np
import pytest
from pydantic import BaseModel, ValidationError

from tesseract_core.runtime.experimental import LazySequence
from tesseract_core.runtime.schema_types import (
    Array,
    ArrayFlags,
    Bool,
    Differentiable,
    Float64,
    Int32,
    Int64,
    is_differentiable,
)


class MyModel(BaseModel):
    array_int: Array[(2, 3), Int64]
    array_float: Differentiable[Array[(None, 3), Float64]]
    array_bool: Array[..., Bool]
    scalar_int: Differentiable[Int32]


arr_int = np.array([[1, 2, 3], [4, 5, 6]])
arr_float = np.array([[1.0, 2.0, 3.0]])
arr_bool = np.array([True, False, True]).reshape(1, 1, 3)
scalar_int = np.int32(42)


def test_generic_array_type_repr():
    typ = Array[(2, 3), Int64]
    assert repr(typ) == "typing.Annotated[numpy.ndarray, Array[(2, 3), 'int64']]"


def test_flag_propagation():
    get_flags = lambda model, field: model.model_fields[field].metadata[0].flags
    assert get_flags(MyModel, "array_int") == ()
    assert get_flags(MyModel, "array_float") == (ArrayFlags.DIFFERENTIABLE,)
    assert get_flags(MyModel, "array_bool") == ()
    assert get_flags(MyModel, "scalar_int") == (ArrayFlags.DIFFERENTIABLE,)


def test_is_differentiable():
    assert is_differentiable(Array[(2, 3), Int64]) is False
    assert is_differentiable(Differentiable[Array[(2, 3), Int64]]) is True
    assert is_differentiable(Differentiable[Int32]) is True
    assert is_differentiable(Int32) is False


def test_model_creation():
    model = MyModel(
        array_int=arr_int,
        array_float=arr_float,
        array_bool=arr_bool,
        scalar_int=scalar_int,
    )

    assert isinstance(model, MyModel)
    assert isinstance(model.array_int, np.ndarray)
    assert isinstance(model.array_float, np.ndarray)
    assert isinstance(model.array_bool, np.ndarray)
    assert isinstance(model.scalar_int, np.int32)
    assert model.array_int.shape == (2, 3)
    assert model.array_float.shape == (1, 3)
    assert model.array_bool.shape == (1, 1, 3)

    assert np.array_equal(model.array_int, arr_int)
    assert np.array_equal(model.array_float, arr_float.reshape(1, 3))
    assert np.array_equal(model.array_bool, arr_bool)
    assert model.scalar_int == scalar_int


def test_model_creation_from_pyobj():
    model = MyModel(
        array_int=arr_int.tolist(),
        array_float=arr_float.tolist(),
        array_bool=arr_bool.tolist(),
        scalar_int=int(scalar_int),
    )

    assert isinstance(model, MyModel)
    assert isinstance(model.array_int, np.ndarray)
    assert isinstance(model.array_float, np.ndarray)
    assert isinstance(model.array_bool, np.ndarray)
    assert isinstance(model.scalar_int, np.int32)
    assert model.array_int.shape == (2, 3)
    assert model.array_float.shape == (1, 3)
    assert model.array_bool.shape == (1, 1, 3)

    assert np.array_equal(model.array_int, arr_int)
    assert np.array_equal(model.array_float, arr_float.reshape(1, 3))
    assert np.array_equal(model.array_bool, arr_bool)
    assert model.scalar_int == scalar_int


def test_json_base64_rountrip():
    model = MyModel(
        array_int=arr_int,
        array_float=arr_float,
        array_bool=arr_bool,
        scalar_int=scalar_int,
    )
    serialized = model.model_dump_json(context={"array_encoding": "base64"})

    assert isinstance(serialized, str)

    try:
        json.loads(serialized)
    except json.JSONDecodeError as exc:
        raise AssertionError("Invalid JSON") from exc

    roundtrip = MyModel.model_validate_json(serialized)

    for field in model.model_fields:
        assert np.array_equal(getattr(roundtrip, field), getattr(model, field))


def test_json_binref_roundtrip(tmpdir):
    dumpdir = Path(tmpdir) / "dumpdir"
    model = MyModel(
        array_int=np.array([[1, 2, 3], [4, 5, 6]]),  # 48 bytes
        array_float=np.array([[1.0, 2.0, 3.0]]),  # 28 bytes
        array_bool=np.array([True, False, True]).reshape(1, 1, 3),  # 3 bytes
        scalar_int=np.int32(42),  # 4 bytes
    )
    serialized = model.model_dump_json(
        context={
            "array_encoding": "binref",
            "base_dir": dumpdir,
            # set file size so that arr_int will be written to one buffer
            # and the remaining arrays to a second buffer
            "max_file_size": 40,
        }
    )

    assert isinstance(serialized, str)
    assert len(list(dumpdir.glob("*.bin"))) == 2

    try:
        json.loads(serialized)
    except json.JSONDecodeError as exc:
        raise AssertionError("Invalid JSON") from exc

    # standard round-trip
    roundtrip = MyModel.model_validate_json(serialized, context={"base_dir": dumpdir})
    for field in model.model_fields:
        assert np.array_equal(getattr(roundtrip, field), getattr(model, field))

    # move all binary files to a different directory and try to load
    # making sure that they contain paths relative to the passed base_dir
    movedir = dumpdir.parent / "movedir"
    dumpdir.rename(movedir)
    roundtrip = MyModel.model_validate_json(serialized, context={"base_dir": movedir})
    for field in model.model_fields:
        assert np.array_equal(getattr(roundtrip, field), getattr(model, field))

    with pytest.raises(ValidationError, match="relative path but no base_dir"):
        model.model_validate_json(serialized)

    with pytest.raises(ValueError, match="base_dir"):
        model.model_dump_json(context={"array_encoding": "binref"})


def test_python_dump_array_encoding():
    model = MyModel(
        array_int=arr_int,
        array_float=arr_float,
        array_bool=arr_bool,
        scalar_int=scalar_int,
    )
    serialized = model.model_dump()
    assert isinstance(serialized["array_int"], np.ndarray)
    assert np.array_equal(serialized["array_int"], arr_int)


def test_python_roundtrip():
    model = MyModel(
        array_int=arr_int,
        array_float=arr_float,
        array_bool=arr_bool,
        scalar_int=scalar_int,
    )
    serialized = model.model_dump()
    roundtrip = MyModel.model_validate(serialized)

    for field in model.model_fields:
        assert np.array_equal(getattr(roundtrip, field), getattr(model, field))


def test_model_from_json(tmpdir):
    payload = {
        "array_int": {
            "shape": [2, 3],
            "dtype": "int64",
            "data": {
                "buffer": base64.b64encode(arr_int.tobytes()).decode(),
                "encoding": "base64",
            },
            # we include a 'object_type' field in all serializations.
            # for backwards compatibility this field is not required, but if
            # given must be Literal["array"].
            "object_type": "array",
        },
        "array_float": {
            "shape": [1, 3],
            "dtype": "float64",
            "data": {
                "buffer": arr_float.flatten().tolist(),
                "encoding": "json",
            },
        },
        "array_bool": {
            "shape": [1, 1, 3],
            "dtype": "bool",
            "data": {
                "buffer": str(
                    tmpdir / "data.bin"
                ),  # test without offset and absolute path
                "encoding": "binref",
            },
        },
        "scalar_int": {
            "shape": [],
            "dtype": "int32",
            "data": {
                "buffer": "data.bin:3",  # test with offset and relative path
                "encoding": "binref",
            },
        },
    }
    with open(tmpdir / "data.bin", "wb") as fi:
        fi.write(arr_bool.tobytes())
        fi.write(scalar_int.tobytes())
    payload_json = json.dumps(payload)

    model = MyModel.model_validate_json(
        payload_json, context={"base_dir": Path(tmpdir)}
    )
    ref_model = MyModel(
        array_int=arr_int,
        array_float=arr_float,
        array_bool=arr_bool,
        scalar_int=scalar_int,
    )

    for field in ref_model.model_fields:
        assert np.array_equal(getattr(model, field), getattr(ref_model, field))


def fix_fake_arrays(fakedata, target_shape, seed=42):
    is_array = (
        lambda x: isinstance(x, dict) and "shape" in x and "dtype" in x and "data" in x
    )
    rng = np.random.RandomState(seed)

    def _walk(data):
        if is_array(data):
            if data["shape"]:  # don't touch scalars
                data["shape"] = target_shape

            new_data = rng.uniform(0, 100, data["shape"]).astype(data["dtype"])

            if data["data"]["encoding"] == "base64":
                data["data"]["buffer"] = base64.b64encode(new_data.tobytes()).decode()
            elif data["data"]["encoding"] == "json":
                data["data"]["buffer"] = new_data.flatten().tolist()
            elif data["data"]["encoding"] == "binref":
                # replace binref with base64 so we don't need to write files
                data["data"]["encoding"] = "base64"
                data["data"]["buffer"] = base64.b64encode(new_data.tobytes()).decode()
        elif isinstance(data, dict):
            for key, value in data.items():
                data[key] = _walk(value)
        elif isinstance(data, list):
            for idx, value in enumerate(data):
                data[idx] = _walk(value)

        return data

    return _walk(fakedata)


@pytest.mark.parametrize("mode", ["validation", "serialization"])
def test_json_schema(mode):
    """Test that array JSON schemas produce valid payloads."""
    schema = MyModel.model_json_schema(mode=mode)

    faker = jsf.JSF(schema)
    payload = faker.generate()

    # Fix fake arrays to match the expected shape and be properly encoded
    # TODO: remove shape argument to have it inferred from the schema once JSF supports it
    # (see https://github.com/ghandic/jsf/issues/118)
    payload = fix_fake_arrays(payload, target_shape=(2, 3))

    try:
        MyModel.model_validate_json(json.dumps(payload))
    except Exception as exc:
        raise AssertionError("Failed to validate generated payload") from exc


def test_json_schema_flags_and_typenames():
    schema = MyModel.model_json_schema()

    array_int_typeref = schema["properties"]["array_int"]["$ref"]
    assert array_int_typeref == "#/$defs/EncodedArrayModel__2_3__int64__noflags"
    array_int_def = schema["$defs"][array_int_typeref.split("/")[-1]]
    assert array_int_def["array_flags"] == []

    array_float_typeref = schema["properties"]["array_float"]["$ref"]
    assert (
        array_float_typeref
        == "#/$defs/EncodedArrayModel__any_3__float64__DIFFERENTIABLE"
    )
    array_float_def = schema["$defs"][array_float_typeref.split("/")[-1]]
    assert array_float_def["array_flags"] == ["DIFFERENTIABLE"]

    array_bool_typeref = schema["properties"]["array_bool"]["$ref"]
    assert array_bool_typeref == "#/$defs/EncodedArrayModel__anyrank__bool__noflags"
    array_bool_def = schema["$defs"][array_bool_typeref.split("/")[-1]]
    assert array_bool_def["array_flags"] == []

    scalar_int_typeref = schema["properties"]["scalar_int"]["$ref"]
    assert (
        scalar_int_typeref == "#/$defs/EncodedArrayModel__scalar__int32__DIFFERENTIABLE"
    )
    scalar_int_def = schema["$defs"][scalar_int_typeref.split("/")[-1]]
    assert scalar_int_def["array_flags"] == ["DIFFERENTIABLE"]


def test_lazy_sequence_array(tmpdir):
    class MyLazyModel(BaseModel):
        array_int_seq: LazySequence[Array[(2, 3), Int64]]

    # Test with in-memory arrays
    model = MyLazyModel(array_int_seq=[arr_int, arr_int])
    assert len(model.array_int_seq) == 2

    for arr in model.array_int_seq:
        assert np.array_equal(arr, arr_int)
        assert arr.shape == (2, 3)
        assert arr.dtype == np.int64

    # Test with files
    encoded_array = {
        "object_type": "array",
        "shape": [2, 3],
        "dtype": "int64",
        "data": {
            "buffer": base64.b64encode(arr_int.tobytes()).decode(),
            "encoding": "base64",
        },
    }

    with open(tmpdir / "inputs_1.json", "w") as f:
        json.dump(encoded_array, f)

    with open(tmpdir / "inputs_2.json", "w") as f:
        json.dump(encoded_array, f)

    model = MyLazyModel(array_int_seq=f"@{tmpdir / 'inputs_*.json'}")
    assert len(model.array_int_seq) == 2
    for arr in model.array_int_seq:
        assert np.array_equal(arr, arr_int)
        assert arr.shape == (2, 3)
        assert arr.dtype == np.int64

    # Test with a file that violates the wrapped schema
    with open(tmpdir / "inputs_3.json", "w") as f:
        json.dump(
            {
                "object_type": "array",
                "shape": [2, 3],
                "dtype": "int64",
                "data": {"buffer": "not a valid buffer", "encoding": "base64"},
            },
            f,
        )

    model = MyLazyModel(array_int_seq=f"@{tmpdir / 'inputs_*.json'}")
    assert len(model.array_int_seq) == 3
    assert np.array_equal(model.array_int_seq[0], arr_int)

    with pytest.raises(ValidationError):
        model.array_int_seq[2]

    # Test serialization
    model = MyLazyModel(array_int_seq=[arr_int, arr_int])
    dumped = model.model_dump(context={"array_encoding": "base64"}, mode="json")
    assert dumped == {"array_int_seq": [encoded_array, encoded_array]}

    dumped = model.model_dump(mode="python")
    assert np.array_equal(dumped["array_int_seq"][0], arr_int)
    assert np.array_equal(dumped["array_int_seq"][1], arr_int)


def test_lazy_sequence_pyobj(tmpdir):
    class MyLazyModel(BaseModel):
        py_seq: LazySequence[tuple[str, int]]

    test_seq = [("a", 1), ("b", 2)]
    model = MyLazyModel(py_seq=test_seq)
    assert list(model.py_seq) == test_seq

    dumped = model.model_dump()
    assert dumped["py_seq"] == test_seq

    # Test with files
    with open(tmpdir / "inputs_1.json", "w") as f:
        json.dump(test_seq[0], f)

    with open(tmpdir / "inputs_2.json", "w") as f:
        json.dump(test_seq[1], f)

    model = MyLazyModel(py_seq=f"@{tmpdir / 'inputs_*.json'}")

    assert list(model.py_seq) == test_seq

    # Test with a file that violates the wrapped schema
    with open(tmpdir / "inputs_3.json", "w") as f:
        json.dump({"key": "value"}, f)

    model = MyLazyModel(py_seq=f"@{tmpdir / 'inputs_*.json'}")
    assert len(model.py_seq) == 3
    assert model.py_seq[0] == test_seq[0]

    with pytest.raises(ValidationError):
        model.py_seq[2]


def test_lazy_sequence_json_schema():
    class MyLazyModel(BaseModel):
        array_int_seq: LazySequence[Array[(2, 3), Int64]]

    schema = MyLazyModel.model_json_schema()

    allowed_types = schema["properties"]["array_int_seq"]["anyOf"]
    assert len(allowed_types) == 2
    assert allowed_types[0]["type"] == "array"
    assert allowed_types[1]["type"] == "string"


def test_dtype_casting():
    json_payload_str = MyModel(
        array_int=arr_int,
        array_float=arr_float,
        array_bool=arr_bool,
        scalar_int=scalar_int,
    ).model_dump_json()

    # Base case: proper int data (should work fine)
    json_payload = json.loads(json_payload_str)
    json_payload["array_int"]["data"] = {
        "buffer": arr_int.flatten().tolist(),
        "encoding": "json",
    }
    res = MyModel.model_validate(json_payload)
    assert np.array_equal(res.array_int, arr_int)

    # Case 1: floats in JSON array w/o fractional parts (should work fine)
    json_payload = json.loads(json_payload_str)
    json_payload["array_int"]["data"] = {
        "buffer": arr_int.astype(float).flatten().tolist(),
        "encoding": "json",
    }
    res = MyModel.model_validate(json_payload)
    assert np.array_equal(res.array_int, arr_int)

    # Case 2: floats in JSON array w/ fractional parts (should raise)
    json_payload = json.loads(json_payload_str)
    json_payload["array_int"]["data"] = {
        "buffer": (arr_int.astype(float) + 1e-6).flatten().tolist(),
        "encoding": "json",
    }
    with pytest.raises(ValidationError, match="Expected integer data"):
        MyModel.model_validate(json_payload)

    # Case 3: pass NumPy array directly (should work fine)
    json_payload = json.loads(json_payload_str)
    json_payload["array_int"] = arr_int
    res = MyModel.model_validate(json_payload)
    assert np.array_equal(res.array_int, arr_int)

    # Case 4: pass NumPy array with incompatible dtype (should raise)
    json_payload = json.loads(json_payload_str)
    json_payload["array_int"] = arr_int.astype(np.float32)
    with pytest.raises(ValidationError, match="cannot be cast"):
        MyModel.model_validate(json_payload)

    # Case 5: pass JSON data directly (should work fine)
    json_payload = json.loads(json_payload_str)
    json_payload["array_int"] = arr_int.tolist()
    res = MyModel.model_validate(json_payload)
    assert np.array_equal(res.array_int, arr_int)

    # Case 6: pass JSON data with incompatible dtype (should raise)
    json_payload = json.loads(json_payload_str)
    json_payload["array_int"] = arr_int.astype(np.float32).tolist()
    with pytest.raises(ValidationError, match="cannot be cast"):
        MyModel.model_validate(json_payload)

    # Case 7: Pass non-numeric data (should raise)
    json_payload = json.loads(json_payload_str)
    json_payload["array_int"] = ["a", "b", "c"]
    with pytest.raises(ValidationError, match="Could not convert object"):
        MyModel.model_validate(json_payload)

    # Case 8: Pass non-numeric Python object (should raise)
    json_payload = json.loads(json_payload_str)
    json_payload["array_int"] = object()
    with pytest.raises(ValidationError, match="Could not convert object"):
        MyModel.model_validate(json_payload)



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import json
import os
import random
import string
from pathlib import Path
from shutil import copytree
from textwrap import indent
from traceback import format_exception
from typing import Any

import pytest

# NOTE: Do NOT import tesseract_core here, as it will cause typeguard to fail

here = Path(__file__).parent

UNIT_TESSERACT_PATH = here / ".." / "examples"
UNIT_TESSERACTS = [Path(tr).stem for tr in UNIT_TESSERACT_PATH.glob("*/")]


def pytest_addoption(parser):
    parser.addoption(
        "--always-run-endtoend",
        action="store_true",
        dest="run_endtoend",
        help="Never skip end-to-end tests",
        default=None,
    )
    parser.addoption(
        "--skip-endtoend",
        action="store_false",
        dest="run_endtoend",
        help="Skip end-to-end tests",
    )
    parser.addoption(
        "--tesseract-dir",
        action="store",
        default=None,
        dest="tesseract_dir",
        help="Directory of your tesseract api",
    )


@pytest.fixture
def tesseract_dir(request):
    """Return the tesseract directory."""
    # This is used to set the tesseract_dir fixture
    # in the pytest_generate_tests function above.
    tesseract_dir = request.config.getoption("tesseract_dir")
    if tesseract_dir:
        return Path(tesseract_dir)
    return None


def pytest_collection_modifyitems(config, items):
    """Ensure that endtoend tests are run last (expensive!)."""
    # Map items to containing directory
    dir_mapping = {item: Path(item.module.__file__).parent.stem for item in items}

    # Sort items based on directory
    sorted_items = sorted(items, key=lambda item: dir_mapping[item] == "endtoend_tests")
    items[:] = sorted_items

    # Add skip marker to endtoend tests if not explicitly enabled
    # or if Docker is not available
    def has_docker():
        from tesseract_core.sdk import docker_client as docker_client_module

        try:
            docker = docker_client_module.CLIDockerClient()
            docker.info()
            return True
        except Exception:
            return False

    run_endtoend = config.getvalue("run_endtoend")

    if run_endtoend is None:
        # tests may be skipped if Docker is not available
        run_endtoend = has_docker()
        skip_reason = "Docker is required for this test"
    elif not run_endtoend:
        skip_reason = "Skipping end-to-end tests"

    if not run_endtoend:
        for item in items:
            if dir_mapping[item] == "endtoend_tests":
                item.add_marker(pytest.mark.skip(reason=skip_reason))


@pytest.fixture(scope="session")
def unit_tesseract_names():
    """Return all unit tesseract names."""
    return UNIT_TESSERACTS


@pytest.fixture(scope="session", params=UNIT_TESSERACTS)
def unit_tesseract_path(request) -> Path:
    """Parametrized fixture to return all unit tesseracts."""
    # pass only tesseract names as params to get prettier test names
    return UNIT_TESSERACT_PATH / request.param


@pytest.fixture()
def dummy_docker_file(tmp_path):
    """Create a dummy Dockerfile for testing."""
    dockerfile_path = tmp_path / "Dockerfile"
    dockerfile_content = """
        FROM alpine

        ENTRYPOINT ["/bin/sh", "-c"]

        # Set environment variables
        ENV TESSERACT_NAME="dummy-tesseract"
        """
    with open(dockerfile_path, "w") as f:
        f.write(dockerfile_content)
    return dockerfile_path


@pytest.fixture(scope="session")
def dummy_tesseract_location():
    """Return the dummy tesseract location."""
    return here / "dummy_tesseract"


@pytest.fixture
def dummy_tesseract_package(tmpdir, dummy_tesseract_location):
    """Create a dummy tesseract package on disk for testing."""
    copytree(dummy_tesseract_location, tmpdir, dirs_exist_ok=True)
    return Path(tmpdir)


@pytest.fixture
def dummy_tesseract_module(dummy_tesseract_package):
    """Create a dummy tesseract module for testing."""
    from tesseract_core.runtime.core import load_module_from_path

    return load_module_from_path(dummy_tesseract_package / "tesseract_api.py")


@pytest.fixture
def dummy_tesseract(dummy_tesseract_package):
    """Set tesseract_api_path env var for testing purposes."""
    from tesseract_core.runtime.config import get_config, update_config

    orig_config_kwargs = {}
    orig_path = get_config().api_path
    # default may have been used and tesseract_api.py is not guaranteed to exist
    # therefore, we only pass the original path in cleanup if not equal to default
    if orig_path != Path("tesseract_api.py"):
        orig_config_kwargs |= {"api_path": orig_path}
    api_path = Path(dummy_tesseract_package / "tesseract_api.py").resolve()

    try:
        # Configure via envvar so we also propagate it to subprocesses
        os.environ["TESSERACT_API_PATH"] = str(api_path)
        update_config(api_path=api_path)
        yield
    finally:
        # As this is used by an auto-use fixture, cleanup may happen
        # after dummy_tesseract_noenv has already unset
        if "TESSERACT_API_PATH" in os.environ:
            del os.environ["TESSERACT_API_PATH"]
        update_config(**orig_config_kwargs)


@pytest.fixture
def dummy_tesseract_noenv(dummy_tesseract_package):
    """Use without tesseract_api_path to test handling of this."""
    from tesseract_core.runtime.config import get_config, update_config

    orig_api_path = get_config().api_path
    orig_cwd = os.getcwd()

    # Ensure TESSERACT_API_PATH is not set with python os
    if "TESSERACT_API_PATH" in os.environ:
        del os.environ["TESSERACT_API_PATH"]

    try:
        os.chdir(dummy_tesseract_package)
        update_config()
        yield
    finally:
        update_config(api_path=orig_api_path)
        os.chdir(orig_cwd)


@pytest.fixture
def free_port():
    """Find a free port to use for HTTP."""
    import socket
    from contextlib import closing

    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("127.0.0.1", 0))
        return s.getsockname()[1]


@pytest.fixture(scope="session")
def docker_client():
    from tesseract_core.sdk import docker_client as docker_client_module

    return docker_client_module.CLIDockerClient()


@pytest.fixture
def docker_volume(docker_client):
    # Create the Docker volume
    volume_name = f"test_volume_{''.join(random.choices(string.ascii_lowercase + string.digits, k=8))}"
    volume = docker_client.volumes.create(name=volume_name)
    try:
        yield volume
    finally:
        volume.remove(force=True)


@pytest.fixture(scope="module")
def docker_cleanup_module(docker_client, request):
    """Clean up all tesseracts created by the tests after the module exits."""
    return _docker_cleanup(docker_client, request)


@pytest.fixture
def docker_cleanup(docker_client, request):
    """Clean up all tesseracts created by the tests after the test exits."""
    return _docker_cleanup(docker_client, request)


def _docker_cleanup(docker_client, request):
    """Clean up all tesseracts created by the tests."""
    # Shared object to track what objects need to be cleaned up in each test
    context = {"images": [], "project_ids": [], "containers": [], "volumes": []}

    def pprint_exc(e: BaseException) -> str:
        """Pretty print exception."""
        return "".join(
            indent(line, "  ") for line in format_exception(type(e), e, e.__traceback__)
        )

    def cleanup_func():
        failures = []

        # Teardown projects first
        for project_id in context["project_ids"]:
            try:
                docker_client.compose.down(project_id)
            except Exception as e:
                failures.append(
                    f"Failed to tear down project {project_id}: {pprint_exc(e)}"
                )

        # Remove containers
        for container in context["containers"]:
            try:
                if isinstance(container, str):
                    container_obj = docker_client.containers.get(container)
                else:
                    container_obj = container

                container_obj.remove(v=True, force=True)
            except Exception as e:
                failures.append(
                    f"Failed to remove container {container}: {pprint_exc(e)}"
                )

        # Remove images
        for image in context["images"]:
            try:
                if isinstance(image, str):
                    image_obj = docker_client.images.get(image)
                else:
                    image_obj = image

                docker_client.images.remove(image_obj.id)
            except Exception as e:
                failures.append(f"Failed to remove image {image}: {pprint_exc(e)}")

        # Remove volumes
        for volume in context["volumes"]:
            try:
                if isinstance(volume, str):
                    volume_obj = docker_client.volumes.get(volume)
                else:
                    volume_obj = volume

                volume_obj.remove(force=True)
            except Exception as e:
                failures.append(f"Failed to remove volume {volume}: {pprint_exc(e)}")

        if failures:
            raise RuntimeError(
                "Failed to clean up some Docker objects during test teardown:\n"
                + "\n".join(failures)
            )

    request.addfinalizer(cleanup_func)
    return context


@pytest.fixture
def dummy_image_name():
    """Create a dummy image name, and clean up after the test."""
    image_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=16))
    image_name = f"tmp_tesseract_image_{image_id}"
    yield image_name


@pytest.fixture(scope="module")
def shared_dummy_image_name():
    """Create a dummy image name, and clean up after all tests."""
    image_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=16))
    image_name = f"tmp_tesseract_image_{image_id}"
    yield image_name


@pytest.fixture
def mocked_docker(monkeypatch):
    """Mock CLIDockerClient class."""
    import tesseract_core.sdk.docker_client
    from tesseract_core.sdk import engine
    from tesseract_core.sdk.docker_client import Container, Image, NotFound

    class MockedContainer(Container):
        """Mock Container class."""

        def __init__(self, return_args: dict):
            self.return_args = return_args

        def wait(self, **kwargs: Any):
            """Mock wait method for Container."""
            return {"StatusCode": 0, "Error": None}

        @property
        def name(self):
            """Mock name property for Container."""
            return json.dumps(self.return_args)

        @property
        def attrs(self):
            """Mock attrs method for Container."""
            return {"Config": {"Env": ["TESSERACT_NAME=vectoradd"]}}

        def logs(self, stderr=False, stdout=False, **kwargs: Any) -> bytes:
            """Mock logs method for Container."""
            res_stdout = json.dumps(self.return_args).encode("utf-8")
            res_stderr = b"hello tesseract"
            if stdout and stderr:
                return res_stdout + res_stderr
            if stdout:
                return res_stdout
            return res_stderr

        def remove(self, **kwargs: Any):
            """Mock remove method for Container."""
            pass

    created_ids = set()

    class MockedDocker:
        """Mock CLIDockerClient class."""

        @staticmethod
        def info() -> tuple:
            """Mock info method for DockerClient."""
            return "", ""

        class volumes:
            """Mock of CLIDockerClient.volumes."""

            @staticmethod
            def create(name: str) -> Any:
                """Mock of CLIDockerClient.volumes.create."""
                return {"Name": name}

            @staticmethod
            def get(name: str) -> Any:
                """Mock of CLIDockerClient.volumes.get."""
                if "/" in name:
                    raise NotFound(f"Volume {name} not found")
                return {"Name": name}

            @staticmethod
            def list() -> list[Any]:
                """Mock of CLIDockerClient.volumes.list."""
                return [{"Name": "test_volume"}]

        class images:
            """Mock of CLIDockerClient.images."""

            @staticmethod
            def get(name: str) -> Image:
                """Mock of CLIDockerClient.images.get."""
                return MockedDocker.images.list()[0]

            @staticmethod
            def list() -> list[Image]:
                """Mock of CLIDockerClient.images.list."""
                return [
                    Image.from_dict(
                        {
                            "Id": "sha256:123456789abcdef",
                            "RepoTags": ["vectoradd:latest"],
                            "Size": 123456789,
                            "Config": {"Env": ["TESSERACT_NAME=vectoradd"]},
                        },
                    ),
                    Image.from_dict(
                        {
                            "Id": "sha256:48932484029303",
                            "RepoTags": ["hello-world:latest"],
                            "Size": 43829489032,
                            "Config": {"Env": ["PATH=/fake-path"]},
                        },
                    ),
                ]

            @staticmethod
            def buildx(*args, **kwargs) -> Image:
                return MockedDocker.images.list()[0]

        class containers:
            @staticmethod
            def get(name: str) -> MockedContainer:
                """Mock of CLIDockerClient.containers.get."""
                if name == "vectoradd":
                    return MockedContainer({"TESSERACT_NAME": "vectoradd"})
                raise NotFound(f"Container {name} not found")

            @staticmethod
            def list() -> list[MockedContainer]:
                """Mock of CLIDockerClient.containers.list."""
                return [MockedContainer({"TESSERACT_NAME": "vectoradd"})]

            @staticmethod
            def run(**kwargs: Any) -> MockedContainer | tuple[bytes, bytes]:
                """Mock run method for containers."""
                container = MockedContainer(kwargs)
                if kwargs.get("detach", False):
                    return container
                return (
                    container.logs(stdout=True, stderr=False),
                    container.logs(stdout=False, stderr=True),
                )

        class compose:
            @staticmethod
            def list() -> set:
                """Return ids of all created tesseracts projects."""
                return created_ids

            @staticmethod
            def up(compose_fpath: str, project_name: str) -> str:
                """Mock of CLIDockerClient.compose.up."""
                created_ids.add(project_name)
                return project_name

            @staticmethod
            def down(project_id: str) -> bool:
                """Mock of CLIDockerClient.compose.down."""
                created_ids.remove(project_id)
                return True

            @staticmethod
            def exists(project_id: str) -> bool:
                """Mock of CLIDockerClient.compose.exists."""
                return project_id in created_ids

    mock_instance = MockedDocker()
    monkeypatch.setattr(engine, "docker_client", mock_instance)
    monkeypatch.setattr(engine, "is_podman", lambda: False)
    monkeypatch.setattr(
        tesseract_core.sdk.docker_client, "CLIDockerClient", MockedDocker
    )

    def hacked_get(url, *args, **kwargs):
        if url.endswith("/health"):
            # Simulate a successful health check
            return type("Response", (), {"status_code": 200, "json": lambda: {}})()
        raise NotImplementedError(f"Mocked get request to {url} not implemented")

    engine.requests.get = hacked_get

    yield mock_instance



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import json
import urllib.parse
from pathlib import Path
from typing import Any, Literal, Optional, Union, get_args

import fsspec
from pydantic import TypeAdapter

PathLike = Union[str, Path]

supported_format_type = Literal["json", "msgpack", "json+base64", "json+binref"]
SUPPORTED_FORMATS = get_args(supported_format_type)


def guess_format_from_path(path: PathLike) -> supported_format_type:
    """Guess the format from the given path.

    The format is determined by the file extension.
    """
    if path.endswith(".json"):
        return "json"
    elif path.endswith(".msgpack"):
        return "msgpack"

    raise ValueError(
        f"Could not guess format from path {path} (supported formats: {SUPPORTED_FORMATS})"
    )


def output_to_bytes(
    obj: Any, format: supported_format_type, base_dir: Optional[Union[str, Path]] = None
) -> bytes:
    """Encode endpoint output to bytes in the given format.

    obj may contain pydantic.BaseModel / RootModel instances, or regular Python objects.
    """
    ObjSchema = TypeAdapter(type(obj))
    if format == "json":
        return ObjSchema.dump_json(
            obj, context={"array_encoding": "json"}, exclude_unset=True
        )
    elif format == "json+base64":
        return ObjSchema.dump_json(
            obj, context={"array_encoding": "base64"}, exclude_unset=True
        )
    elif format == "json+binref":
        return ObjSchema.dump_json(
            obj,
            context={"array_encoding": "binref", "base_dir": base_dir},
            exclude_unset=True,
        )
    elif format == "msgpack":
        import msgpack

        import tesseract_core.runtime.vendor.msgpack_numpy as msgpack_numpy

        obj_clean = ObjSchema.dump_python(obj, exclude_unset=True)
        return msgpack.packb(obj_clean, default=msgpack_numpy.encode)

    raise ValueError(
        f"Unsupported format {format} (must be one of {SUPPORTED_FORMATS})"
    )


def load_bytes(
    buffer: bytes,
    format: supported_format_type,
) -> Any:
    """Decode the given buffer to a Python object.

    The buffer is expected to be in the given format.
    """
    if format.startswith("json"):
        return json.loads(buffer.decode())
    elif format == "msgpack":
        import msgpack

        import tesseract_core.runtime.vendor.msgpack_numpy as msgpack_numpy

        return msgpack.unpackb(buffer, object_hook=msgpack_numpy.decode)

    raise ValueError(
        f"Unsupported format {format} (must be one of {SUPPORTED_FORMATS})"
    )


def read_from_path(path: PathLike, offset: int = 0, length: int = -1) -> bytes:
    """Read the contents of the given path as bytes.

    Path may be anything supported by fsspec.
    """
    with fsspec.open(path, "rb") as f:
        if offset != 0:
            f.seek(offset)
        return f.read(length)


def write_to_path(buffer: bytes, path: PathLike, append: bool = False) -> None:
    """Write the buffer to the given path.

    Path may be anything supported by fsspec.
    """
    mode = "ab" if append else "wb"
    with fsspec.open(path, mode, auto_mkdir=True) as f:
        f.write(buffer)


def expand_glob(pattern: str) -> list[str]:
    """Expand the given glob pattern.

    Path may be anything supported by fsspec.
    """
    open_files = fsspec.open_files(pattern, "rb", expand=True)
    return sorted(f.path for f in open_files)


def get_filesize(path: PathLike) -> int:
    """Get the size of the given path in bytes.

    Path may be anything supported by fsspec.
    """
    try:
        with fsspec.open(path, "rb") as f:
            f.seek(0, 2)
            return f.tell()
    except FileNotFoundError:
        return 0


def join_paths(base: PathLike, other: PathLike) -> str:
    """Join the base path (URL or local path) with the given other path.

    If the other path is an absolute URL, return it as is.
    """
    if is_absolute_path(other):
        return str(other)
    if is_url(base):
        return urllib.parse.urljoin(base, other)
    return Path(base).joinpath(other).as_posix()


def is_absolute_path(path: PathLike) -> bool:
    """Check if path is an absolute path or a url."""
    return is_url(path) or Path(path).is_absolute()


def is_url(path: PathLike) -> bool:
    """Check if path is a url."""
    return bool(urllib.parse.urlparse(str(path)).scheme)


def parent_path(x: PathLike) -> PathLike:
    """Get parent of given path (which may be a URL)."""
    if is_url(x):
        return urllib.parse.urljoin(x, "..")

    return type(x)(Path(x).parent.as_posix())



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import atexit
import base64
import traceback
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from functools import cached_property, wraps
from pathlib import Path
from types import ModuleType
from typing import Any
from urllib.parse import urlparse, urlunparse

import numpy as np
import requests
from pydantic import BaseModel, TypeAdapter, ValidationError
from pydantic_core import InitErrorDetails

from tesseract_core.runtime.config import update_config

from . import engine

PathLike = str | Path


@dataclass
class SpawnConfig:
    """Configuration for spawning a Tesseract."""

    image: str
    volumes: list[str] | None
    environment: dict[str, str] | None
    gpus: list[str] | None
    num_workers: int
    debug: bool
    no_compose: bool


def requires_client(func: Callable) -> Callable:
    """Decorator to require a client for a Tesseract instance."""

    @wraps(func)
    def wrapper(self: Tesseract, *args: Any, **kwargs: Any) -> Any:
        if not self._client:
            raise RuntimeError(
                f"{self.__class__.__name__} must be used as a context manager when created via `from_image`."
            )
        return func(self, *args, **kwargs)

    return wrapper


class Tesseract:
    """A Tesseract.

    This class represents a single Tesseract instance, either remote or local,
    and provides methods to run commands on it and retrieve results.

    Communication between a Tesseract and this class is done either via
    HTTP requests or directly via Python calls to the Tesseract API.
    """

    def __init__(self, url: str) -> None:
        self._spawn_config = None
        self._serve_context = None
        self._lastlog = None
        self._client = HTTPClient(url)

    @classmethod
    def from_url(cls, url: str) -> Tesseract:
        """Create a Tesseract instance from a URL.

        This is useful for connecting to a remote Tesseract instance.

        Args:
            url: The URL of the Tesseract instance.

        Returns:
            A Tesseract instance.
        """
        obj = cls.__new__(cls)
        obj.__init__(url)
        return obj

    @classmethod
    def from_image(
        cls,
        image: str,
        *,
        volumes: list[str] | None = None,
        environment: dict[str, str] | None = None,
        input_path: PathLike | None = None,
        output_path: PathLike | None = None,
        gpus: list[str] | None = None,
        num_workers: int = 1,
        no_compose: bool = False,
    ) -> Tesseract:
        """Create a Tesseract instance from a Docker image.

        When using this method, the Tesseract will be spawned in a Docker
        container, serving the Tesseract API via HTTP. To use the Tesseract,
        you need to call the `serve` method or use it as a context manager.

        Example:
            >>> with Tesseract.from_image("my_tesseract") as t:
            ...    # Use tesseract here

        This will automatically teardown the Tesseract when exiting the
        context manager.

        Args:
            image: The Docker image to use.
            volumes: List of volumes to mount, e.g. ["/path/on/host:/path/in/container"].
            environment: dictionary of environment variables to pass to the Tesseract.
            input_path: Path to be mounted as the input directory in the
                container (read only). All paths in the input payload must be
                relative to this path.
            output_path: Path to be mounted as the output directory in the
                container (read+write). All paths in the output result will be
                relative to this path.
            gpus: List of GPUs to use, e.g. ["0", "1"]. (default: no GPUs)
            num_workers: Number of worker processes to use. This determines how
                many requests can be handled in parallel. Higher values
                will increase throughput, but also increase resource usage.
            no_compose: if True, do not use Docker Compose to serve the Tesseracts.

        Returns:
            A Tesseract instance.
        """
        obj = cls.__new__(cls)

        if environment is None:
            environment = {}
        if volumes is None:
            volumes = []
        if input_path is not None:
            input_path = Path(input_path).resolve()
            volumes.append(f"{input_path}:/tesseract/input_data:ro")
        if output_path is not None:
            output_path = Path(output_path).resolve()
            volumes.append(f"{output_path}:/tesseract/output_data:rw")

        obj._spawn_config = SpawnConfig(
            image=image,
            volumes=volumes,
            environment=environment,
            gpus=gpus,
            num_workers=num_workers,
            debug=True,
            no_compose=no_compose,
        )
        obj._serve_context = None
        obj._lastlog = None
        obj._client = None
        return obj

    @classmethod
    def from_tesseract_api(
        cls,
        tesseract_api: str | Path | ModuleType,
        input_path: Path | None = None,
        output_path: Path | None = None,
    ) -> Tesseract:
        """Create a Tesseract instance from a Tesseract API module.

        Warning: This does not use a containerized Tesseract, but rather
        imports the Tesseract API directly. This is useful for debugging,
        but requires a matching runtime environment + all dependencies to be
        installed locally.

        Args:
            tesseract_api: Path to the `tesseract_api.py` file, or an
                already imported Tesseract API module.
            input_path: Path of input directory. All paths in the tesseract
                payload have to be relative to this path.
            output_path: Path of output directory. All paths in the tesseract
                result with be given relative to this path.

        Returns:
            A Tesseract instance.
        """
        if isinstance(tesseract_api, str | Path):
            from tesseract_core.runtime.core import load_module_from_path

            tesseract_api_path = Path(tesseract_api).resolve(strict=True)
            if not tesseract_api_path.is_file():
                raise RuntimeError(
                    f"Tesseract API path {tesseract_api_path} is not a file."
                )

            try:
                tesseract_api = load_module_from_path(tesseract_api_path)
            except ImportError as ex:
                raise RuntimeError(
                    f"Cannot load Tesseract API from {tesseract_api_path}"
                ) from ex

        if input_path is not None:
            update_config(input_path=str(input_path.resolve()))
        if output_path is not None:
            update_config(output_path=str(output_path.resolve()))

        obj = cls.__new__(cls)
        obj._spawn_config = None
        obj._serve_context = None
        obj._lastlog = None
        obj._client = LocalClient(tesseract_api)
        return obj

    def __enter__(self) -> Tesseract:
        """Enter the Tesseract context.

        This will start the Tesseract server if it is not already running.
        """
        if self._serve_context is not None:
            raise RuntimeError("Cannot serve the same Tesseract multiple times.")

        if self._client is not None:
            # Tesseract is already being served -> no-op
            return self

        self.serve()
        return self

    def __exit__(self, *args: Any) -> None:
        """Exit the Tesseract context.

        This will stop the Tesseract server if it is running.
        """
        if self._serve_context is None:
            # This can happen if __enter__ short-cirtuits
            return
        self.teardown()

    def server_logs(self) -> str:
        """Get the logs of the Tesseract server.

        Returns:
            logs of the Tesseract server.
        """
        if self._spawn_config is None:
            raise RuntimeError(
                "Can only retrieve logs for a Tesseract created via from_image."
            )
        if self._serve_context is None:
            return self._lastlog
        return engine.logs(self._serve_context["container_id"])

    def serve(self, port: str | None = None, host_ip: str = "127.0.0.1") -> None:
        """Serve the Tesseract.

        Args:
            port: Port to serve the Tesseract on.
            host_ip: IP address of the host to bind the Tesseract to.
        """
        if self._spawn_config is None:
            raise RuntimeError("Can only serve a Tesseract created via from_image.")
        if self._serve_context is not None:
            raise RuntimeError("Tesseract is already being served.")
        project_id, container_id, served_port = self._serve(
            self._spawn_config.image,
            port=port,
            volumes=self._spawn_config.volumes,
            environment=self._spawn_config.environment,
            gpus=self._spawn_config.gpus,
            num_workers=self._spawn_config.num_workers,
            debug=self._spawn_config.debug,
            no_compose=self._spawn_config.no_compose,
            host_ip=host_ip,
        )
        self._serve_context = dict(
            project_id=project_id,
            container_id=container_id,
            port=served_port,
        )
        self._lastlog = None
        self._client = HTTPClient(f"http://{host_ip}:{served_port}")
        atexit.register(self.teardown)

    def teardown(self) -> None:
        """Teardown the Tesseract.

        This will stop and remove the Tesseract container.
        """
        if self._serve_context is None:
            raise RuntimeError("Tesseract is not being served.")
        self._lastlog = self.server_logs()
        engine.teardown(self._serve_context["project_id"])
        self._client = None
        self._serve_context = None
        atexit.unregister(self.teardown)

    def __del__(self) -> None:
        """Destructor for the Tesseract class.

        This will teardown the Tesseract if it is being served.
        """
        if self._serve_context is not None:
            self.teardown()

    @staticmethod
    def _serve(
        image: str,
        port: str | None = None,
        host_ip: str = "127.0.0.1",
        volumes: list[str] | None = None,
        environment: dict[str, str] | None = None,
        gpus: list[str] | None = None,
        debug: bool = False,
        num_workers: int = 1,
        no_compose: bool = False,
    ) -> tuple[str, str, int]:
        if port is not None:
            ports = [port]
        else:
            ports = None

        project_id = engine.serve(
            [image],
            ports=ports,
            volumes=volumes,
            environment=environment,
            gpus=gpus,
            debug=debug,
            num_workers=num_workers,
            host_ip=host_ip,
            no_compose=no_compose,
        )

        first_container = engine.get_project_containers(project_id)[0]
        return project_id, first_container.id, int(first_container.host_port)

    @cached_property
    @requires_client
    def openapi_schema(self) -> dict:
        """Get the OpenAPI schema of this Tessseract.

        Returns:
            dictionary with the OpenAPI Schema.
        """
        return self._client.run_tesseract("openapi_schema")

    @cached_property
    @requires_client
    def input_schema(self) -> dict:
        """Get the input schema of this Tessseract.

        Returns:
             dictionary with the input schema.
        """
        return self._client.run_tesseract("input_schema")

    @cached_property
    @requires_client
    def output_schema(self) -> dict:
        """Get the output schema of this Tessseract.

        Returns:
             dictionary with the output schema.
        """
        return self._client.run_tesseract("output_schema")

    @property
    @requires_client
    def available_endpoints(self) -> list[str]:
        """Get the list of available endpoints.

        Returns:
            a list with all available endpoints for this Tesseract.
        """
        return [endpoint.lstrip("/") for endpoint in self.openapi_schema["paths"]]

    @requires_client
    def apply(self, inputs: dict) -> dict:
        """Run apply endpoint.

        Args:
            inputs: a dictionary with the inputs.

        Returns:
            dictionary with the results.
        """
        payload = {"inputs": inputs}
        return self._client.run_tesseract("apply", payload)

    @requires_client
    def abstract_eval(self, abstract_inputs: dict) -> dict:
        """Run abstract eval endpoint.

        Args:
            abstract_inputs: a dictionary with the (abstract) inputs.

        Returns:
            dictionary with the results.
        """
        payload = {"inputs": abstract_inputs}
        return self._client.run_tesseract("abstract_eval", payload)

    @requires_client
    def health(self) -> dict:
        """Check the health of the Tesseract.

        Returns:
            dictionary with the health status.
        """
        return self._client.run_tesseract("health")

    @requires_client
    def jacobian(
        self, inputs: dict, jac_inputs: list[str], jac_outputs: list[str]
    ) -> dict:
        """Calculate the Jacobian of (some of the) outputs w.r.t. (some of the) inputs.

        Args:
            inputs: a dictionary with the inputs.
            jac_inputs: Inputs with respect to which derivatives will be calculated.
            jac_outputs: Outputs which will be differentiated.

        Returns:
            dictionary with the results.
        """
        if "jacobian" not in self.available_endpoints:
            raise NotImplementedError("Jacobian not implemented for this Tesseract.")

        payload = {
            "inputs": inputs,
            "jac_inputs": jac_inputs,
            "jac_outputs": jac_outputs,
        }
        return self._client.run_tesseract("jacobian", payload)

    @requires_client
    def jacobian_vector_product(
        self,
        inputs: dict,
        jvp_inputs: list[str],
        jvp_outputs: list[str],
        tangent_vector: dict,
    ) -> dict:
        """Calculate the Jacobian Vector Product (JVP) of (some of the) outputs w.r.t. (some of the) inputs.

        Args:
            inputs: a dictionary with the inputs.
            jvp_inputs: Inputs with respect to which derivatives will be calculated.
            jvp_outputs: Outputs which will be differentiated.
            tangent_vector: Element of the tangent space to multiply with the Jacobian.

        Returns:
            dictionary with the results.
        """
        if "jacobian_vector_product" not in self.available_endpoints:
            raise NotImplementedError(
                "Jacobian Vector Product (JVP) not implemented for this Tesseract."
            )

        payload = {
            "inputs": inputs,
            "jvp_inputs": jvp_inputs,
            "jvp_outputs": jvp_outputs,
            "tangent_vector": tangent_vector,
        }
        return self._client.run_tesseract("jacobian_vector_product", payload)

    @requires_client
    def vector_jacobian_product(
        self,
        inputs: dict,
        vjp_inputs: list[str],
        vjp_outputs: list[str],
        cotangent_vector: dict,
    ) -> dict:
        """Calculate the Vector Jacobian Product (VJP) of (some of the) outputs w.r.t. (some of the) inputs.

        Args:
            inputs: a dictionary with the inputs.
            vjp_inputs: Inputs with respect to which derivatives will be calculated.
            vjp_outputs: Outputs which will be differentiated.
            cotangent_vector: Element of the cotangent space to multiply with the Jacobian.


        Returns:
            dictionary with the results.
        """
        if "vector_jacobian_product" not in self.available_endpoints:
            raise NotImplementedError(
                "Vector Jacobian Product (VJP) not implemented for this Tesseract."
            )

        payload = {
            "inputs": inputs,
            "vjp_inputs": vjp_inputs,
            "vjp_outputs": vjp_outputs,
            "cotangent_vector": cotangent_vector,
        }
        return self._client.run_tesseract("vector_jacobian_product", payload)


def _tree_map(func: Callable, tree: Any, is_leaf: Callable | None = None) -> Any:
    """Recursively apply a function to all leaves of a tree-like structure."""
    if is_leaf is not None and is_leaf(tree):
        return func(tree)
    if isinstance(tree, Mapping):  # Dictionary-like structure
        return {key: _tree_map(func, value, is_leaf) for key, value in tree.items()}

    if isinstance(tree, Sequence) and not isinstance(
        tree, (str, bytes)
    ):  # List, tuple, etc.
        return type(tree)(_tree_map(func, item, is_leaf) for item in tree)

    # If nothing above matched do nothing
    return tree


def _encode_array(arr: np.ndarray, b64: bool = True) -> dict:
    if b64:
        data = {
            "buffer": base64.b64encode(arr.tobytes()).decode(),
            "encoding": "base64",
        }
    else:
        data = {
            "buffer": arr.tolist(),
            "encoding": "raw",
        }

    return {
        "shape": arr.shape,
        "dtype": arr.dtype.name,
        "data": data,
    }


def _decode_array(encoded_arr: dict) -> np.ndarray:
    if "data" in encoded_arr:
        if encoded_arr["data"]["encoding"] == "base64":
            data = base64.b64decode(encoded_arr["data"]["buffer"])
            arr = np.frombuffer(data, dtype=encoded_arr["dtype"])
        else:
            arr = np.array(encoded_arr["data"]["buffer"], dtype=encoded_arr["dtype"])
    else:
        raise ValueError("Encoded array does not contain 'data' key. Cannot decode.")
    arr = arr.reshape(encoded_arr["shape"])
    return arr


class HTTPClient:
    """HTTP Client for Tesseracts."""

    def __init__(self, url: str) -> None:
        self._url = self._sanitize_url(url)

    @staticmethod
    def _sanitize_url(url: str) -> str:
        parsed = urlparse(url)

        if not parsed.scheme:
            url = f"http://{url}"
            parsed = urlparse(url)

        sanitized = urlunparse((parsed.scheme, parsed.netloc, parsed.path, "", "", ""))
        return sanitized

    @property
    def url(self) -> str:
        """(Sanitized) URL to connect to."""
        return self._url

    def _request(
        self, endpoint: str, method: str = "GET", payload: dict | None = None
    ) -> dict:
        url = f"{self.url}/{endpoint.lstrip('/')}"

        if payload:
            encoded_payload = _tree_map(
                _encode_array, payload, is_leaf=lambda x: hasattr(x, "shape")
            )
        else:
            encoded_payload = None

        response = requests.request(method=method, url=url, json=encoded_payload)

        if response.status_code == requests.codes.unprocessable_entity:
            # Try and raise a more helpful error if the response is a Pydantic error
            try:
                data = response.json()
            except requests.JSONDecodeError:
                # Is not a Pydantic error
                data = {}
            if "detail" in data:
                errors = []
                for e in data["detail"]:
                    ctx = e.get("ctx", {})
                    if not ctx.get("error") and e.get("msg"):
                        # Hacky, but msg contains info like "Value error, ...",
                        # which will be prepended to the message anyway by pydantic.
                        # This way, we remove whatever is before the first comma.
                        msg = e["msg"].partition(", ")[2]
                        ctx["error"] = msg

                    error = InitErrorDetails(
                        type=e["type"],
                        loc=tuple(e["loc"]),
                        input=e.get("input"),
                        ctx=ctx,
                    )
                    errors.append(error)

                raise ValidationError.from_exception_data(
                    f"endpoint {endpoint}", line_errors=errors
                )

        if not response.ok:
            raise RuntimeError(
                f"Error {response.status_code} from Tesseract: {response.text}"
            )

        data = response.json()

        if endpoint in [
            "apply",
            "jacobian",
            "jacobian_vector_product",
            "vector_jacobian_product",
        ]:
            data = _tree_map(
                _decode_array,
                data,
                is_leaf=lambda x: type(x) is dict and "shape" in x,
            )

        return data

    def run_tesseract(self, endpoint: str, payload: dict | None = None) -> dict:
        """Run a Tesseract endpoint.

        Args:
            endpoint: The endpoint to run.
            payload: The payload to send to the endpoint.

        Returns:
            The loaded JSON response from the endpoint, with decoded arrays.
        """
        if endpoint in [
            "input_schema",
            "output_schema",
            "openapi_schema",
            "health",
        ]:
            method = "GET"
        else:
            method = "POST"

        if endpoint == "openapi_schema":
            endpoint = "openapi.json"

        return self._request(endpoint, method, payload)


class LocalClient:
    """Local Client for Tesseracts."""

    def __init__(self, tesseract_api: ModuleType) -> None:
        from tesseract_core.runtime.core import create_endpoints
        from tesseract_core.runtime.serve import create_rest_api

        self._endpoints = {
            func.__name__: func for func in create_endpoints(tesseract_api)
        }
        self._openapi_schema = create_rest_api(tesseract_api).openapi()

    def run_tesseract(self, endpoint: str, payload: dict | None = None) -> dict:
        """Run a Tesseract endpoint.

        Args:
            endpoint: The endpoint to run.
            payload: The payload to send to the endpoint.

        Returns:
            The loaded JSON response from the endpoint, with decoded arrays.
        """
        if endpoint == "openapi_schema":
            return self._openapi_schema

        if endpoint not in self._endpoints:
            raise RuntimeError(f"Endpoint {endpoint} not found in Tesseract API.")

        func = self._endpoints[endpoint]
        InputSchema = func.__annotations__.get("payload", None)
        OutputSchema = func.__annotations__.get("return", None)

        if InputSchema is not None:
            parsed_payload = InputSchema.model_validate(payload)
        else:
            parsed_payload = None

        try:
            if parsed_payload is not None:
                result = self._endpoints[endpoint](parsed_payload)
            else:
                result = self._endpoints[endpoint]()
        except Exception as ex:
            # Some clients like Tesseract-JAX swallow tracebacks from re-raised exceptions, so we explicitly
            # format the traceback here to include it in the error message.
            tb = traceback.format_exc()
            raise RuntimeError(
                f"{tb}\nError running Tesseract API {endpoint}: {ex} (see above for full traceback)"
            ) from None

        if OutputSchema is not None:
            # Validate via schema, then dump to stay consistent with other clients
            if isinstance(OutputSchema, type) and issubclass(OutputSchema, BaseModel):
                result = OutputSchema.model_validate(result).model_dump()
            else:
                result = TypeAdapter(OutputSchema).validate_python(result)

        return result



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import json
import os

from pydantic import BaseModel

#
# Schemas
#


class InputSchema(BaseModel):
    pass


class OutputSchema(BaseModel):
    pass


#
# Required endpoints
#


def apply(inputs: InputSchema) -> OutputSchema:
    dirname = os.path.dirname(__file__)
    parameter_file = os.path.join(dirname, "parameters.json")

    # check file exists
    assert os.path.exists(parameter_file)

    # check parameters
    with open(parameter_file) as file:
        data = json.load(file)

    assert data == {"a": 1.0, "b": 100.0}

    return OutputSchema()



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Instantiated version of the FastAPI app."""

from tesseract_core.runtime.core import get_tesseract_api
from tesseract_core.runtime.serve import create_rest_api

tesseract_api = get_tesseract_api()
app = create_rest_api(tesseract_api)



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from pydantic import BaseModel


class InputSchema(BaseModel):
    pass


class OutputSchema(BaseModel):
    pass


def apply(inputs: InputSchema) -> OutputSchema:
    return OutputSchema()



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Tests for validation logic in tesseract_runtime/core.py.

Test a couple of edge cases and failure modes regarding schema validation and AD inputs / outputs.
"""

from collections.abc import Sequence
from types import ModuleType

import numpy as np
import pytest
from pydantic import BaseModel, ValidationError

from tesseract_core.runtime import Array, Differentiable, Float32
from tesseract_core.runtime.core import create_endpoints
from tesseract_core.runtime.tree_transforms import get_at_path


@pytest.fixture
def testmodule():
    class _InputSchema(BaseModel):
        array_seq: Sequence[Differentiable[Array[(3, 1), Float32]]]
        array_dict: dict[str, Differentiable[Array[(None,), Float32]]]
        scalar: Float32
        scalar_diff: Differentiable[Float32]

    class _OutputSchema(BaseModel):
        result_seq: Sequence[Differentiable[Array[(10,), Float32]]]
        result_arr: Array[(None,), Float32]

    class FakeModule(ModuleType):
        @property
        def InputSchema(self):
            return _InputSchema

        @property
        def OutputSchema(self):
            return _OutputSchema

        def apply(self, inputs):
            res_seq = [np.ones(10)] * 3
            res_float = np.zeros(5)
            return _OutputSchema(result_seq=res_seq, result_arr=res_float)

        def jacobian(self, inputs, jac_inputs, jac_outputs):
            out = {}
            for out_key in jac_outputs:
                out_shape = (10,)
                out[out_key] = {}
                for in_key in jac_inputs:
                    if in_key.startswith("array_seq"):
                        container, idx = in_key.split(".")
                        idx = int(idx[1:-1])
                    elif in_key.startswith("array_dict"):
                        container, idx = in_key.split(".")
                        idx = idx[1:-1]
                    else:
                        container = in_key
                        idx = None

                    if container == "array_seq":
                        out[out_key][in_key] = np.ones(
                            (*out_shape, *inputs.array_seq[idx].shape)
                        )
                    elif container == "array_dict":
                        out[out_key][in_key] = np.ones(
                            (*out_shape, *inputs.array_dict[idx].shape)
                        )
                    else:
                        out[out_key][in_key] = np.ones(out_shape)
            return out

        def jacobian_vector_product(
            self, inputs, jvp_inputs, jvp_outputs, tangent_vector
        ):
            out = {}
            for out_key in jvp_outputs:
                out_shape = (10,)
                out[out_key] = np.ones(out_shape)
            return out

        def vector_jacobian_product(
            self, inputs, vjp_inputs, vjp_outputs, cotangent_vector
        ):
            out = {}
            for in_key in vjp_inputs:
                if in_key.startswith("array_seq"):
                    container, idx = in_key.split(".")
                    idx = int(idx[1:-1])
                elif in_key.startswith("array_dict"):
                    container, idx = in_key.split(".")
                    idx = idx[1:-1]
                else:
                    container = in_key
                    idx = None

                if container == "array_seq":
                    out[in_key] = np.ones(inputs.array_seq[idx].shape)
                elif container == "array_dict":
                    out[in_key] = np.ones(inputs.array_dict[idx].shape)
                else:
                    out[in_key] = np.ones(())

            return out

        def abstract_eval(self, abstract_inputs):
            return {
                "result_seq": [{"shape": (10,), "dtype": "float32"}] * 3,
                "result_arr": {"shape": (5,), "dtype": "float32"},
            }

    return FakeModule("TestModule")


test_input = {
    "array_seq": [
        np.array([1.0, 2.0, 3.0]).reshape(3, 1),
        np.array([4.0, 5.0, 6.0]).reshape(3, 1),
    ],
    "array_dict": {"a": np.array([7.0, 8.0, 9.0, 10.0]), "b": np.array([11.0, 12.0])},
    "scalar": 1.0,
    "scalar_diff": np.array(2.0),
}
test_output = {
    "result_seq": [np.ones(10)] * 3,
    "result_arr": np.zeros(5),
}


def _recurse_pytree(pytree, func):
    if isinstance(pytree, dict):
        return {k: _recurse_pytree(v, func) for k, v in pytree.items()}
    elif isinstance(pytree, list | tuple):
        return [_recurse_pytree(v, func) for v in pytree]
    else:
        return func(pytree)


def _find_endpoint(endpoint_list, endpoint_name):
    for endpoint in endpoint_list:
        if endpoint.__name__ == endpoint_name:
            if "payload" in endpoint.__annotations__:
                schema = endpoint.__annotations__["payload"]
            else:
                schema = None
            return endpoint, schema
    raise ValueError(f"Endpoint {endpoint_name} not found.")


def test_create_endpoints(testmodule):
    endpoints = create_endpoints(testmodule)
    assert endpoints


def test_apply_endpoint(testmodule):
    endpoints = create_endpoints(testmodule)

    apply_endpoint, EndpointSchema = _find_endpoint(endpoints, "apply")
    assert apply_endpoint.__name__ == "apply"

    inputs = EndpointSchema.model_validate({"inputs": test_input})
    result = apply_endpoint(inputs)
    np.testing.assert_equal(result.model_dump(), test_output)


def test_apply_returns_dict(testmodule):
    # Mock the apply method to return a dictionary
    def apply(inputs):
        return test_output

    testmodule.apply = apply

    endpoints = create_endpoints(testmodule)

    apply_endpoint, EndpointSchema = _find_endpoint(endpoints, "apply")
    assert apply_endpoint.__name__ == "apply"

    inputs = EndpointSchema.model_validate({"inputs": test_input})
    result = apply_endpoint(inputs)
    np.testing.assert_equal(result.model_dump(), test_output)


def test_abstract_eval_endpoint(testmodule):
    endpoints = create_endpoints(testmodule)

    abstract_eval_endpoint, EndpointSchema = _find_endpoint(endpoints, "abstract_eval")
    assert abstract_eval_endpoint.__name__ == "abstract_eval"

    def _array_to_shapedtype(obj):
        obj = np.asarray(obj)
        return {"shape": obj.shape, "dtype": str(obj.dtype)}

    inputs_shapedtype = _recurse_pytree(test_input, _array_to_shapedtype)
    inputs = EndpointSchema.model_validate({"inputs": inputs_shapedtype})
    result = abstract_eval_endpoint(inputs)
    assert result

    with pytest.raises(ValueError):
        failing_inputs_shapedtype = inputs_shapedtype.copy()
        failing_inputs_shapedtype["scalar"]["shape"] = (1, 2, 3)
        inputs = EndpointSchema.model_validate({"inputs": failing_inputs_shapedtype})


def test_schemas_contain_diffable_paths_extra(testmodule):
    endpoints = create_endpoints(testmodule)
    input_schema_endpoint, _ = _find_endpoint(endpoints, "input_schema")

    result = input_schema_endpoint()
    assert "differentiable_arrays" in result
    assert result["differentiable_arrays"].keys() == {
        "array_dict.{}",
        "array_seq.[]",
        "scalar_diff",
    }

    output_schema_endpoint, _ = _find_endpoint(endpoints, "output_schema")
    result = output_schema_endpoint()
    assert "differentiable_arrays" in result
    assert result["differentiable_arrays"].keys() == {"result_seq.[]"}


@pytest.mark.parametrize(
    "ad_inout",
    [
        # Valid combinations for jac_inputs and jac_outputs
        ({"array_seq.[0]", "array_dict.{a}", "scalar_diff"}, {"result_seq.[0]"}),
        ({"array_seq.[0]", "array_dict.{a}"}, {"result_seq.[0]"}),
        ({"array_seq.[0]", "array_dict.{a}", "scalar_diff"}, {"result_seq.[0]"}),
        ({"array_seq.[0]", "array_dict.{a}"}, {"result_seq.[0]"}),
        ({"array_seq.[0]", "array_seq.[1]"}, {"result_seq.[0]"}),
    ],
)
@pytest.mark.parametrize(
    "endpoint_name", ["jacobian", "jacobian_vector_product", "vector_jacobian_product"]
)
def test_ad_endpoint(testmodule, ad_inout, endpoint_name):
    endpoints = create_endpoints(testmodule)

    endpoint_func, EndpointSchema = _find_endpoint(endpoints, endpoint_name)
    assert endpoint_func.__name__ == endpoint_name

    ad_inp, ad_out = ad_inout

    if endpoint_name == "jacobian":
        inputs = {"inputs": test_input, "jac_inputs": ad_inp, "jac_outputs": ad_out}
    elif endpoint_name == "jacobian_vector_product":
        tangent_vector = {k: get_at_path(test_input, k) for k in ad_inp}
        inputs = {
            "inputs": test_input,
            "jvp_inputs": ad_inp,
            "jvp_outputs": ad_out,
            "tangent_vector": tangent_vector,
        }
    elif endpoint_name == "vector_jacobian_product":
        cotangent_vector = {k: get_at_path(test_output, k) for k in ad_out}
        inputs = {
            "inputs": test_input,
            "vjp_inputs": ad_inp,
            "vjp_outputs": ad_out,
            "cotangent_vector": cotangent_vector,
        }

    inputs = EndpointSchema.model_validate(inputs)
    result = endpoint_func(inputs)
    assert result


@pytest.mark.parametrize(
    "ad_inout_invalid",
    [
        # -- Invalid jac_inputs, valid jac_outputs --
        # Unknown keys
        (
            "String should match pattern",
            {"array_seq.[0]", "array_dict.{a}", "scalar_diff", "invalid_key"},
            {"result_seq.[0]"},
        ),
        # Invalid container index
        (
            "String should match pattern",
            {"array_seq.[foobar]", "array_dict.{a}", "scalar_diff", "invalid_key"},
            {"result_seq.[0]"},
        ),
        # Non-existent container index
        (
            "array_seq.[100]",
            {"array_seq.[100]", "array_dict.{a}", "scalar_diff"},
            {"result_seq.[0]"},
        ),
        (
            "array_dict.{xyz}",
            {"array_seq.[0]", "array_dict.{xyz}", "scalar_diff"},
            {"result_seq.[0]"},
        ),
        # Non-differentiable keys
        (
            "String should match pattern",
            {"array_seq.[0]", "array_dict.{a}", "scalar"},
            {"result_seq.[0]"},
        ),
        # Container without index
        (
            "String should match pattern",
            {"array_seq", "array_dict.{a}", "scalar_diff"},
            {"result_seq.[0]"},
        ),
        (
            "String should match pattern",
            {"array_seq.[0]", "array_dict", "scalar_diff"},
            {"result_seq.[0]"},
        ),
        # Missing keys
        ("at least 1 item", set(), {"result_seq.[0]"}),
        # -- Valid jac_inputs, invalid jac_outputs --
        # Unknown keys
        (
            "String should match pattern",
            {"array_seq.[0]", "array_dict.{a}", "scalar_diff"},
            {"result_seq.[0]", "invalid_key"},
        ),
        # Non-differentiable keys
        (
            "String should match pattern",
            {"array_seq.[0]", "array_dict.{a}", "scalar_diff"},
            {"result_seq.[0]", "result_arr"},
        ),
        # Container without index
        (
            "String should match pattern",
            {"array_seq.[0]", "array_dict.{a}", "scalar_diff"},
            {"result_seq"},
        ),
        # Missing keys
        ("at least 1 item", {"scalar_diff"}, set()),
    ],
)
@pytest.mark.parametrize(
    "endpoint_name", ["jacobian", "jacobian_vector_product", "vector_jacobian_product"]
)
def test_ad_endpoint_invalid(testmodule, ad_inout_invalid, endpoint_name):
    endpoints = create_endpoints(testmodule)

    endpoint_func, EndpointSchema = _find_endpoint(endpoints, endpoint_name)
    assert endpoint_func.__name__ == endpoint_name

    msg, ad_inp, ad_out = ad_inout_invalid

    if endpoint_name == "jacobian":
        inputs = {"inputs": test_input, "jac_inputs": ad_inp, "jac_outputs": ad_out}
    elif endpoint_name == "jacobian_vector_product":
        tangent_vector = {}
        for k in ad_inp:
            try:
                tangent_vector[k] = get_at_path(test_input, k)
            except (KeyError, ValueError, IndexError):
                # Path is invalid, add a dummy value
                tangent_vector[k] = np.ones(10)
        inputs = {
            "inputs": test_input,
            "jvp_inputs": ad_inp,
            "jvp_outputs": ad_out,
            "tangent_vector": tangent_vector,
        }
    elif endpoint_name == "vector_jacobian_product":
        cotangent_vector = {}
        for k in ad_out:
            try:
                cotangent_vector[k] = get_at_path(test_output, k)
            except (KeyError, ValueError, IndexError):
                # Path is invalid, add a dummy value
                cotangent_vector[k] = np.ones(10)
        inputs = {
            "inputs": test_input,
            "vjp_inputs": ad_inp,
            "vjp_outputs": ad_out,
            "cotangent_vector": cotangent_vector,
        }

    with pytest.raises(ValidationError) as excinfo:
        inputs = EndpointSchema.model_validate(inputs)
        endpoint_func(inputs)

    assert msg in str(excinfo.value)


@pytest.mark.parametrize(
    "endpoint_name", ["jacobian_vector_product", "vector_jacobian_product"]
)
@pytest.mark.parametrize("failure_mode", ["missing", "extra", "invalid"])
def test_ad_endpoint_bad_tangent(testmodule, endpoint_name, failure_mode):
    endpoints = create_endpoints(testmodule)

    endpoint_func, EndpointSchema = _find_endpoint(endpoints, endpoint_name)
    assert endpoint_func.__name__ == endpoint_name

    ad_inp = {"array_seq.[0]", "array_dict.{a}", "scalar_diff"}
    ad_out = {"result_seq.[0]"}

    if endpoint_name == "jacobian_vector_product":
        if failure_mode == "missing":
            tangent_vector = {}
            msg = "Expected tangent vector with keys"
        elif failure_mode == "extra":
            tangent_vector = {k: get_at_path(test_input, k) for k in ad_inp}
            tangent_vector["invalid_key"] = np.ones(10)
            msg = "String should match pattern"
        elif failure_mode == "invalid":
            tangent_vector = {k: "ahoy" for k in ad_inp}
            msg = "Could not convert object"

        inputs = {
            "inputs": test_input,
            "jvp_inputs": ad_inp,
            "jvp_outputs": ad_out,
            "tangent_vector": tangent_vector,
        }

    elif endpoint_name == "vector_jacobian_product":
        if failure_mode == "missing":
            cotangent_vector = {}
            msg = "Expected cotangent vector with keys"
        elif failure_mode == "extra":
            cotangent_vector = {k: get_at_path(test_output, k) for k in ad_out}
            cotangent_vector["invalid_key"] = np.ones(10)
            msg = "String should match pattern"
        elif failure_mode == "invalid":
            cotangent_vector = {k: "ahoy" for k in ad_out}
            msg = "Could not convert object"

        inputs = {
            "inputs": test_input,
            "vjp_inputs": ad_inp,
            "vjp_outputs": ad_out,
            "cotangent_vector": cotangent_vector,
        }

    with pytest.raises(ValidationError, match=msg):
        inputs = EndpointSchema.model_validate(inputs)
        endpoint_func(inputs)



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""End-to-end tests for all unit Tesseracts.

Each unit Tesseract is tested by building a Docker image from it and then
running the Tesseract CLI + HTTP interface to test its functionality.

Add test cases for specific unit Tesseracts to the TEST_CASES dictionary.
"""

import base64
import json
import re
import time
import traceback
from dataclasses import dataclass
from pathlib import Path

import numpy as np
import numpy.typing as npt
import pytest
import requests
from common import build_tesseract, image_exists
from typer.testing import CliRunner


# Only necessary when matching multi-word string
def format_stderr(stderr: str) -> str:
    no_color = re.sub(r"\x1b\[[0-9;]*m", "", stderr)
    return " ".join(re.sub(r"[^\w \d_.,!?:;\-]+", " ", no_color).split())


def json_normalize(obj: str):
    """Normalize JSON str for comparison."""
    return json.dumps(json.loads(obj), separators=(",", ":"))


def encode_array(arr, as_json=False):
    """Helper function to encode a numpy array into Tesseract-friendly format."""
    arr = np.asarray(arr)
    out = {
        "object_type": "array",
        "shape": arr.shape,
        "dtype": arr.dtype.name,
        "data": {
            "buffer": base64.b64encode(arr.tobytes()).decode(),
            "encoding": "base64",
        },
    }
    if as_json:
        return json.dumps(out, separators=(",", ":"))

    return out


def assert_contains_array_allclose(
    output: dict | list,
    array_expected: npt.ArrayLike,
    rtol=1e-4,
    atol=1e-4,
):
    """Check the output pytree for an array close to array_expected."""

    def _get_array_leaves(tree):
        if isinstance(tree, dict) and tree.get("object_type") == "array":
            shape = tree["shape"]
            dtype = tree["dtype"]
            data = tree["data"]
            if data["encoding"] == "base64":
                buffer = base64.b64decode(data["buffer"])
                array = np.frombuffer(buffer, dtype=dtype).reshape(shape)
            else:
                raise NotImplementedError("only base64 encoding supported")
            yield array
        elif isinstance(tree, dict):
            for val in tree.values():
                yield from _get_array_leaves(val)
        elif isinstance(tree, list):
            for val in tree:
                yield from _get_array_leaves(val)
        else:
            raise AssertionError(f"Got unexpected type {type(tree)}")

    output_arrays = list(_get_array_leaves(output))
    expected = np.asarray(array_expected)
    for array in output_arrays:
        if np.allclose(array, expected, rtol=rtol, atol=atol):
            return

    raise AssertionError(
        f"Expected array not found in output.\nExpected: {expected}\nFound arrays: {output_arrays}"
    )


@dataclass
class SampleRequest:
    endpoint: str
    payload: dict
    expected_status_code: int = 200
    output_contains_pattern: str | list[str] | None = None
    output_contains_array: npt.ArrayLike | None = None
    output_format: str = "json+base64"


@dataclass
class Config:
    test_with_random_inputs: bool = False
    sample_requests: list[SampleRequest] = None
    volume_mounts: list[str] = None
    input_path: str = None
    output_path: str = None


# Add config and test cases for specific unit Tesseracts here
TEST_CASES = {
    "empty": Config(
        test_with_random_inputs=True,
        sample_requests=[
            SampleRequest(endpoint="apply", payload={"inputs": {}}),
        ],
    ),
    "py39": Config(
        test_with_random_inputs=True,
    ),
    "helloworld": Config(
        test_with_random_inputs=True,
        sample_requests=[
            SampleRequest(
                endpoint="apply",
                payload={"inputs": {"name": "Ozzy"}},
                output_contains_pattern="Hello Ozzy!",
            ),
        ],
    ),
    "pyvista-arm64": Config(
        test_with_random_inputs=True,
        sample_requests=[
            SampleRequest(
                endpoint="apply",
                payload={
                    "inputs": {
                        "vertices": encode_array(
                            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]
                        )
                    }
                },
                output_contains_pattern="pv_mesh",
            ),
        ],
    ),
    "localpackage": Config(
        test_with_random_inputs=True,
        sample_requests=[
            SampleRequest(
                endpoint="apply",
                payload={"inputs": {"name": "Ozzy"}},
                output_contains_pattern="Hello Ozzy!",
            ),
        ],
    ),
    "vectoradd": Config(
        test_with_random_inputs=True,
        sample_requests=[
            SampleRequest(
                endpoint="apply",
                payload={
                    "inputs": {
                        "a": encode_array([1, 2, 3]),
                        "b": encode_array([4, 5, 6]),
                    },
                },
                output_contains_pattern=[encode_array([5.0, 7.0, 9.0], as_json=True)],
            ),
            SampleRequest(
                endpoint="jacobian",
                payload={
                    "inputs": {
                        "a": encode_array([1, 2, 3]),
                        "b": encode_array([4, 5, 6]),
                    },
                    "jac_inputs": ["s", "a"],
                    "jac_outputs": ["result"],
                },
                output_contains_pattern=['"s":', '"a":'],
            ),
            SampleRequest(
                endpoint="jacobian",
                payload={
                    "inputs": {
                        "a": encode_array([1, 2, 3]),
                        "b": encode_array([4, 5, 6]),
                    },
                    "jac_inputs": ["invalid", "inputs"],
                    "jac_outputs": ["result"],
                },
                expected_status_code=422,
                output_contains_pattern="Input should be",
            ),
            SampleRequest(
                endpoint="jacobian",
                payload={
                    "inputs": {
                        "a": encode_array([1, 2, 3]),
                        "b": encode_array([4, 5, 6]),
                    },
                    "jac_inputs": ["s", "a"],
                    "jac_outputs": ["invalid", "outputs"],
                },
                expected_status_code=422,
                output_contains_pattern="Input should be",
            ),
            SampleRequest(
                endpoint="check-gradients",
                payload={
                    "inputs": {
                        "a": encode_array([1, 2, 3]),
                        "b": encode_array([4, 5, 6]),
                    },
                },
            ),
        ],
    ),
    "vectoradd_jax": Config(
        test_with_random_inputs=True,
        sample_requests=[
            SampleRequest(
                endpoint="apply",
                payload={
                    "inputs": {
                        "a": {"v": encode_array([1, 2, 3]), "s": 3},
                        "b": {"v": encode_array([4, 5, 6]), "s": 1},
                    },
                },
                output_contains_array=np.array([7.0, 11.0, 15.0], dtype="float32"),
            ),
            SampleRequest(
                endpoint="abstract_eval",
                payload={
                    "inputs": {
                        "a": {
                            "v": {"shape": [3], "dtype": "float32"},
                            "s": {"shape": [], "dtype": "float32"},
                        },
                        "b": {
                            "v": {"shape": [3], "dtype": "float32"},
                            "s": {"shape": [], "dtype": "float32"},
                        },
                    }
                },
            ),
            SampleRequest(
                endpoint="apply",
                payload={"inputs": {}},
                expected_status_code=422,
                output_contains_pattern="missing",
            ),
            SampleRequest(
                endpoint="jacobian",
                payload={
                    "inputs": {
                        "a": {"v": encode_array([1, 2, 3]), "s": 2},
                        "b": {"v": encode_array([4, 5, 6]), "s": 1},
                    },
                    "jac_inputs": ["a.s", "a.v"],
                    "jac_outputs": ["vector_add.result"],
                },
                output_contains_pattern=['"a.s":', '"a.v":'],
                output_contains_array=np.array([[1, 2, 3]], dtype="float32"),
            ),
            SampleRequest(
                endpoint="jacobian_vector_product",
                payload={
                    "inputs": {
                        "a": {"v": encode_array([1, 2, 3]), "s": 2},
                        "b": {"v": encode_array([4, 5, 6]), "s": 1},
                    },
                    "jvp_inputs": ["a.v"],
                    "jvp_outputs": ["vector_add.result"],
                    "tangent_vector": {"a.v": encode_array([0.1, 0.2, 0.3])},
                },
                output_contains_array=np.array([0.2, 0.4, 0.6], dtype="float32"),
            ),
            SampleRequest(
                endpoint="jacobian_vector_product",
                payload={
                    "inputs": {
                        "a": {"v": encode_array([1, 2, 3]), "s": 2},
                        "b": {"v": encode_array([4, 5, 6]), "s": 1},
                    },
                    "jvp_inputs": ["a.s"],
                    "jvp_outputs": ["vector_add.result"],
                    "tangent_vector": {"a.s": 0.5},
                },
                output_contains_array=np.array([0.5, 1.0, 1.5], dtype="float32"),
            ),
            SampleRequest(
                endpoint="vector_jacobian_product",
                payload={
                    "inputs": {
                        "a": {"v": encode_array([1.0, 2.0, 3.0]), "s": 2},
                        "b": {"v": encode_array([4.0, 5.0, 6.0]), "s": 1},
                    },
                    "vjp_inputs": ["a.v"],
                    "vjp_outputs": ["vector_add.result"],
                    "cotangent_vector": {
                        "vector_add.result": encode_array([0.1, 0.2, 0.3]),
                    },
                },
                output_contains_array=np.array([0.2, 0.4, 0.6], dtype="float32"),
            ),
            SampleRequest(
                endpoint="vector_jacobian_product",
                payload={
                    "inputs": {
                        "a": {"v": encode_array([1.0, 2.0, 3.0]), "s": 2},
                        "b": {"v": encode_array([4.0, 5.0, 6.0]), "s": 1},
                    },
                    "vjp_inputs": ["a.s"],
                    "vjp_outputs": ["vector_add.result"],
                    "cotangent_vector": {
                        "vector_add.result": encode_array([0.1, 0.2, 0.3]),
                    },
                },
                output_contains_array=np.array([1.4], dtype="float32"),
            ),
            SampleRequest(
                endpoint="check-gradients",
                payload={
                    "inputs": {
                        "a": {"v": encode_array([1, 2, 3]), "s": 2},
                        "b": {"v": encode_array([4, 5, 6]), "s": 1},
                    },
                },
            ),
        ],
    ),
    "vectoradd_torch": Config(
        test_with_random_inputs=True,
        sample_requests=[
            SampleRequest(
                endpoint="apply",
                payload={
                    "inputs": {
                        "a": {"v": encode_array([1, 2, 3]), "s": 3},
                        "b": {"v": encode_array([4, 5, 6]), "s": 1},
                    },
                },
                output_contains_array=np.array([7.0, 11.0, 15.0], dtype="float32"),
            ),
            SampleRequest(
                endpoint="apply",
                payload={"inputs": {}},
                expected_status_code=422,
                output_contains_pattern="missing",
            ),
            SampleRequest(
                endpoint="jacobian",
                payload={
                    "inputs": {
                        "a": {"v": encode_array([1, 2, 3]), "s": 2},
                        "b": {"v": encode_array([4, 5, 6]), "s": 1},
                    },
                    "jac_inputs": ["a.s", "a.v"],
                    "jac_outputs": ["vector_add.result"],
                },
                output_contains_pattern=['"a.s":', '"a.v":'],
                output_contains_array=np.array([[1, 2, 3]], dtype="float32"),
            ),
            SampleRequest(
                endpoint="jacobian_vector_product",
                payload={
                    "inputs": {
                        "a": {"v": encode_array([1, 2, 3]), "s": 2},
                        "b": {"v": encode_array([4, 5, 6]), "s": 1},
                    },
                    "jvp_inputs": ["a.v"],
                    "jvp_outputs": ["vector_add.result"],
                    "tangent_vector": {"a.v": encode_array([0.1, 0.2, 0.3])},
                },
                output_contains_array=np.array([0.2, 0.4, 0.6], dtype="float32"),
            ),
            SampleRequest(
                endpoint="jacobian_vector_product",
                payload={
                    "inputs": {
                        "a": {"v": encode_array([1, 2, 3]), "s": 2},
                        "b": {"v": encode_array([4, 5, 6]), "s": 1},
                    },
                    "jvp_inputs": ["a.s"],
                    "jvp_outputs": ["vector_add.result"],
                    "tangent_vector": {"a.s": 0.5},
                },
                output_contains_array=np.array([0.5, 1.0, 1.5], dtype="float32"),
            ),
            SampleRequest(
                endpoint="vector_jacobian_product",
                payload={
                    "inputs": {
                        "a": {"v": encode_array([1.0, 2.0, 3.0]), "s": 2},
                        "b": {"v": encode_array([4.0, 5.0, 6.0]), "s": 1},
                    },
                    "vjp_inputs": ["a.v"],
                    "vjp_outputs": ["vector_add.result"],
                    "cotangent_vector": {
                        "vector_add.result": encode_array([0.1, 0.2, 0.3]),
                    },
                },
                output_contains_array=np.array([0.2, 0.4, 0.6], dtype="float32"),
            ),
            SampleRequest(
                endpoint="vector_jacobian_product",
                payload={
                    "inputs": {
                        "a": {"v": encode_array([1.0, 2.0, 3.0]), "s": 2},
                        "b": {"v": encode_array([4.0, 5.0, 6.0]), "s": 1},
                    },
                    "vjp_inputs": ["a.s"],
                    "vjp_outputs": ["vector_add.result"],
                    "cotangent_vector": {
                        "vector_add.result": encode_array([0.1, 0.2, 0.3]),
                    },
                },
                output_contains_array=np.array([1.4], dtype="float32"),
            ),
        ],
    ),
    "univariate": Config(
        test_with_random_inputs=True,
        sample_requests=[
            SampleRequest(
                endpoint="apply",
                payload={
                    "inputs": {"x": 0.0, "y": 0.0},
                },
                output_contains_array=np.array([1.0], dtype="float64"),
            ),
            SampleRequest(
                endpoint="jacobian",
                payload={
                    "inputs": {"x": 0.0, "y": 0.0},
                    "jac_inputs": ["x"],
                    "jac_outputs": ["result"],
                },
                output_contains_pattern=[
                    f'"result":{{"x":{encode_array(np.float32(-2.0), as_json=True)}}}'
                ],
            ),
            SampleRequest(
                endpoint="jacobian",
                payload={
                    "inputs": {"x": 0.0, "y": 0.0},
                    "jac_inputs": ["y"],
                    "jac_outputs": ["result"],
                },
                output_contains_pattern=[
                    f'"result":{{"y":{encode_array(np.float32(0.0), as_json=True)}}}'
                ],
            ),
            SampleRequest(
                endpoint="jacobian",
                payload={
                    "inputs": {"x": 0.0, "y": 0.0},
                    "jac_inputs": ["hey"],
                    "jac_outputs": ["result"],
                },
                expected_status_code=422,
                output_contains_pattern="Input should be",
            ),
            SampleRequest(
                endpoint="jacobian_vector_product",
                payload={
                    "inputs": {"x": 0.0, "y": 0.0},
                    "jvp_inputs": ["x", "y"],
                    "jvp_outputs": ["result"],
                    "tangent_vector": {"x": 1.0, "y": 0.0},
                },
                output_contains_pattern=f'"result":{encode_array(np.float32(-2.0), as_json=True)}',
            ),
            SampleRequest(
                endpoint="vector_jacobian_product",
                payload={
                    "inputs": {"x": 0.0, "y": 0.0},
                    "vjp_inputs": ["x", "y"],
                    "vjp_outputs": ["result"],
                    "cotangent_vector": {"result": 1.0},
                },
                output_contains_pattern=[
                    f'"x":{encode_array(np.float32(-2.0), as_json=True)}',
                    f'"y":{encode_array(np.float32(0.0), as_json=True)}',
                ],
            ),
            SampleRequest(
                endpoint="check-gradients",
                payload={
                    "inputs": {"x": 0.0, "y": 0.0},
                },
            ),
        ],
    ),
    "package_data": Config(
        test_with_random_inputs=True,
    ),
    "cuda": Config(
        test_with_random_inputs=True,
    ),
    "meshstats": Config(
        test_with_random_inputs=False,
        sample_requests=[
            SampleRequest(
                endpoint="apply",
                payload={
                    "inputs": {
                        "mesh": {
                            "n_points": 5,
                            "n_cells": 2,
                            "points": encode_array(
                                [
                                    [0.0, 666.0, 0.0],
                                    [1.0, 0.0, 0.0],
                                    [0.0, 1.0, 0.0],
                                    [1.0, 1.0, 0.0],
                                    [0.5, 0.5, 1.0],
                                ]
                            ),
                            "num_points_per_cell": encode_array([4, 4]),
                            "cell_connectivity": encode_array([0, 1, 2, 3, 1, 2, 3, 4]),
                            "cell_data": {
                                "temperature": encode_array(
                                    [[100.0, 105.0], [110.0, 115.0]]
                                ),
                                "pressure": encode_array([[1.0, 1.2], [1.1, 1.3]]),
                            },
                            "point_data": {
                                "displacement": encode_array(
                                    [
                                        [0.0, 0.1, 0.2],
                                        [0.1, 0.0, 0.2],
                                        [0.2, 0.1, 0.0],
                                        [0.1, 0.2, 0.1],
                                        [0.2, 0.1, 0.1],
                                    ]
                                ),
                                "velocity": encode_array(
                                    [
                                        [0.0, 0.0, 0.0],
                                        [0.1, 0.0, 0.0],
                                        [0.0, 0.1, 0.0],
                                        [0.0, 0.0, 0.1],
                                        [0.1, 0.1, 0.1],
                                    ]
                                ),
                            },
                        }
                    },
                },
                output_contains_pattern=encode_array(
                    np.array([0.0, 666.0, 0.0], dtype="float32"), as_json=True
                ),
            ),
            SampleRequest(
                endpoint="abstract_eval",
                payload={
                    "inputs": {
                        "mesh": {
                            "n_points": 5,
                            "n_cells": 8,
                            "points": {
                                "shape": [5, 3],
                                "dtype": "float32",
                            },
                            "num_points_per_cell": {
                                "shape": [2],
                                "dtype": "int32",
                            },
                            "cell_connectivity": {
                                "shape": [8],
                                "dtype": "int32",
                            },
                            "cell_data": {
                                "temperature": {
                                    "shape": [2, 2],
                                    "dtype": "float32",
                                },
                                "pressure": {
                                    "shape": [2, 2],
                                    "dtype": "float32",
                                },
                            },
                            "point_data": {
                                "displacement": {
                                    "shape": [5, 3],
                                    "dtype": "float32",
                                },
                                "velocity": {
                                    "shape": [5, 3],
                                    "dtype": "float32",
                                },
                            },
                        }
                    },
                },
                output_contains_pattern='"shape":[3],"dtype":"float32"',
            ),
            SampleRequest(
                endpoint="jacobian",
                payload={
                    "jac_inputs": ["mesh.points"],
                    "jac_outputs": ["statistics.barycenter"],
                    "inputs": {
                        "mesh": {
                            "n_points": 5,
                            "n_cells": 2,
                            "points": encode_array(
                                [
                                    [0.0, 666.0, 0.0],
                                    [1.0, 0.0, 0.0],
                                    [0.0, 1.0, 0.0],
                                    [1.0, 1.0, 0.0],
                                    [0.5, 0.5, 1.0],
                                ]
                            ),
                            "num_points_per_cell": encode_array([4, 4]),
                            "cell_connectivity": encode_array([0, 1, 2, 3, 1, 2, 3, 4]),
                            "cell_data": {
                                "temperature": encode_array(
                                    [[100.0, 105.0], [110.0, 115.0]]
                                ),
                                "pressure": encode_array([[1.0, 1.2], [1.1, 1.3]]),
                            },
                            "point_data": {
                                "displacement": encode_array(
                                    [
                                        [0.0, 0.1, 0.2],
                                        [0.1, 0.0, 0.2],
                                        [0.2, 0.1, 0.0],
                                        [0.1, 0.2, 0.1],
                                        [0.2, 0.1, 0.1],
                                    ]
                                ),
                                "velocity": encode_array(
                                    [
                                        [0.0, 0.0, 0.0],
                                        [0.1, 0.0, 0.0],
                                        [0.0, 0.1, 0.0],
                                        [0.0, 0.0, 0.1],
                                        [0.1, 0.1, 0.1],
                                    ]
                                ),
                            },
                        }
                    },
                },
                output_contains_pattern='"shape":[3,5,3]',
            ),
            SampleRequest(
                endpoint="check-gradients",
                payload={
                    "inputs": {
                        "mesh": {
                            "n_points": 5,
                            "n_cells": 2,
                            "points": encode_array(
                                [
                                    [0.0, 2.0, 0.0],
                                    [1.0, 0.0, 0.0],
                                    [0.0, 1.0, 0.0],
                                    [1.0, 1.0, 0.0],
                                    [0.5, 0.5, 1.0],
                                ]
                            ),
                            "num_points_per_cell": encode_array([4, 4]),
                            "cell_connectivity": encode_array([0, 1, 2, 3, 1, 2, 3, 4]),
                            "cell_data": {
                                "temperature": encode_array(
                                    [[100.0, 105.0], [110.0, 115.0]]
                                ),
                                "pressure": encode_array([[1.0, 1.2], [1.1, 1.3]]),
                            },
                            "point_data": {
                                "displacement": encode_array(
                                    [
                                        [0.0, 0.1, 0.2],
                                        [0.1, 0.0, 0.2],
                                        [0.2, 0.1, 0.0],
                                        [0.1, 0.2, 0.1],
                                        [0.2, 0.1, 0.1],
                                    ]
                                ),
                                "velocity": encode_array(
                                    [
                                        [0.0, 0.0, 0.0],
                                        [0.1, 0.0, 0.0],
                                        [0.0, 0.1, 0.0],
                                        [0.0, 0.0, 0.1],
                                        [0.1, 0.1, 0.1],
                                    ]
                                ),
                            },
                        }
                    },
                },
            ),
        ],
    ),
    "dataloader": Config(
        test_with_random_inputs=False,
        sample_requests=[
            SampleRequest(
                endpoint="apply",
                payload={
                    "inputs": {
                        "data": "@/tesseract/input_data/sample_*.json",
                    },
                },
                output_contains_pattern=[
                    '{"data":[{"object_type":"array","shape":[3,3],"dtype":"float32","data":{"buffer":',
                ],
            ),
            SampleRequest(
                endpoint="check-gradients",
                payload={
                    "inputs": {
                        "data": "@/tesseract/input_data/sample_*.json",
                    },
                },
            ),
        ],
        volume_mounts=["testdata:/tesseract/input_data:ro"],
    ),
    "conda": Config(
        test_with_random_inputs=False,
        sample_requests=[
            SampleRequest(
                endpoint="apply",
                payload={"inputs": {"message": "Hey!"}},
                output_contains_pattern=[r'{"cowsays":"  ____\n| Hey! |\n  ====\n'],
            )
        ],
    ),
    "filereference": Config(
        test_with_random_inputs=False,
        sample_requests=[
            SampleRequest(
                endpoint="apply",
                payload={
                    "inputs": {
                        "data": [
                            "sample_7.json",
                            "sample_6.json",
                            "sample_1.json",
                            "sample_0.json",
                            "sample_3.json",
                            "sample_2.json",
                            "sample_9.json",
                            "sample_5.json",
                            "sample_4.json",
                            "sample_8.json",
                        ]
                    }
                },
                output_contains_pattern=["sample_0.copy"],
            )
        ],
        input_path="testdata",
        output_path="output",
    ),
    "metrics": Config(
        test_with_random_inputs=True,
        sample_requests=[
            SampleRequest(
                endpoint="apply",
                payload={"inputs": {}},
                # Just verify it runs without error - output will be empty
            ),
        ],
    ),
}


@pytest.fixture
def unit_tesseract_config(unit_tesseract_names, unit_tesseract_path):
    for tesseract in TEST_CASES:
        if tesseract not in unit_tesseract_names:
            raise ValueError(
                f"A test case in TEST_CASES refers to a nonexistent Tesseract {tesseract}."
            )

    if unit_tesseract_path.name not in TEST_CASES:
        raise ValueError(
            f"No test case found for Tesseract {unit_tesseract_path.name}."
        )

    return TEST_CASES[unit_tesseract_path.name]


def print_debug_info(result):
    """Print debug info from result of a CLI command if it failed."""
    if result.exit_code == 0:
        return
    print(result.output)
    if result.exc_info:
        traceback.print_exception(*result.exc_info)


def fix_fake_arrays(fakedata, seed=42):
    is_array = (
        lambda x: isinstance(x, dict) and "shape" in x and "dtype" in x and "data" in x
    )
    rng = np.random.RandomState(seed)

    def _walk(data):
        if is_array(data):
            # Use broadcasting to minimize surface for shape errors
            data["shape"] = (1,) * len(data["shape"])

            new_data = rng.uniform(0, 100, data["shape"]).astype(data["dtype"])

            if data["data"]["encoding"] == "base64":
                data["data"]["buffer"] = base64.b64encode(new_data.tobytes()).decode()
            elif data["data"]["encoding"] == "json":
                data["data"]["buffer"] = new_data.flatten().tolist()
            elif data["data"]["encoding"] == "binref":
                # FIXME: overriding with base64 to not have to create files
                # in the tesseract
                data["data"]["encoding"] = "base64"
                data["data"]["buffer"] = base64.b64encode(new_data.tobytes()).decode()
        elif isinstance(data, dict):
            for key, value in data.items():
                data[key] = _walk(value)
        elif isinstance(data, list):
            for idx, value in enumerate(data):
                data[idx] = _walk(value)

        return data

    return _walk(fakedata)


def example_from_json_schema(schema):
    """Generate a random example JSON object from a JSON schema."""
    import jsf

    faker = jsf.JSF(schema)
    payload = faker.generate()
    payload = fix_fake_arrays(payload)
    return payload


def test_unit_tesseract_endtoend(
    docker_client,
    dummy_image_name,
    unit_tesseract_path,
    unit_tesseract_config,
    free_port,
    docker_cleanup,
):
    """Test that unit Tesseract images can be built and used to serve REST API."""
    from tesseract_core.sdk.cli import app

    cli_runner = CliRunner(mix_stderr=False)

    # Stage 1: Build
    img_name = build_tesseract(
        docker_client,
        unit_tesseract_path,
        dummy_image_name,
        tag="sometag",
    )
    assert image_exists(docker_client, img_name)
    docker_cleanup["images"].append(img_name)

    # Stage 2: Test CLI usage
    result = cli_runner.invoke(
        app,
        [
            "run",
            img_name,
            "input-schema",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.output
    input_schema = result.output

    mount_args, io_args = [], []

    if unit_tesseract_config.volume_mounts:
        for mnt in unit_tesseract_config.volume_mounts:
            # Assume that the mount is relative to the Tesseract path
            local_path, *other = mnt.split(":")
            local_path = Path(local_path)
            if not local_path.is_absolute():
                local_path = unit_tesseract_path / local_path
            mnt = ":".join([str(local_path), *other])
            mount_args.extend(["--volume", mnt])

    if unit_tesseract_config.input_path:
        io_args.extend(
            [
                "--input-path",
                str(unit_tesseract_path / unit_tesseract_config.input_path),
            ]
        )
    if unit_tesseract_config.output_path:
        io_args.extend(
            [
                "--output-path",
                str(unit_tesseract_path / unit_tesseract_config.output_path),
            ]
        )

    if unit_tesseract_config.test_with_random_inputs:
        random_input = example_from_json_schema(json.loads(input_schema))

        result = cli_runner.invoke(
            app,
            [
                "run",
                img_name,
                *mount_args,
                "apply",
                json.dumps(random_input),
            ],
            catch_exceptions=False,
        )
        assert result.exit_code == 0, result.output

    if unit_tesseract_config.sample_requests:
        for request in unit_tesseract_config.sample_requests:
            print(f"Running request: {request}")
            cli_cmd = request.endpoint.replace("_", "-")

            if cli_cmd == "check-gradients":
                args = [
                    "run",
                    img_name,
                    cli_cmd,
                    json.dumps(request.payload),
                ]
            else:
                args = [
                    "run",
                    img_name,
                    *mount_args,
                    cli_cmd,
                    json.dumps(request.payload),
                    *io_args,
                    "--output-format",
                    request.output_format,
                ]

            result = cli_runner.invoke(app, args)
            if request.expected_status_code == 200:
                print_debug_info(result)
                assert result.exit_code == 0, result.exception
                if cli_cmd in ("check-gradients",):
                    # Result is text
                    output = result.output
                elif unit_tesseract_config.output_path:
                    with open(
                        unit_tesseract_path
                        / unit_tesseract_config.output_path
                        / "results.json"
                    ) as fi:
                        output = json_normalize(fi.read())
                else:
                    # Result is JSON output
                    output = json_normalize(result.output)
            else:
                assert result.exit_code != 0
                # Result is an error message
                output = "".join(traceback.format_exception(*result.exc_info))
                # Click with rich adds color codes and boxes to the output
                # which we need to remove in case they break multi-word pattern matching
                output = format_stderr(output)

            if request.output_contains_pattern is not None:
                patterns = request.output_contains_pattern
                if isinstance(patterns, str):
                    patterns = [patterns]

                for pattern in patterns:
                    assert pattern in output

            if request.output_contains_array is not None:
                array = request.output_contains_array
                output_json = json.loads(output)
                assert_contains_array_allclose(output_json, array)

    # Stage 3: Test HTTP server
    if unit_tesseract_config.volume_mounts is not None:
        # TODO: Mounts are not supported in HTTP mode yet, skip rest of the test for now
        return

    # Cannot mix stderr if we want to load the json
    cli_runner = CliRunner(mix_stderr=False)
    project_id = None

    run_res = cli_runner.invoke(
        app,
        [
            "serve",
            img_name,
            "-p",
            free_port,
        ],
        catch_exceptions=False,
    )

    assert run_res.exit_code == 0, run_res.stderr
    assert run_res.stdout

    project_meta = json.loads(run_res.stdout)
    project_id = project_meta["project_id"]
    docker_cleanup["project_ids"].append(project_id)

    # Give server some time to start up
    timeout = 10
    interval = 0.1
    while True:
        try:
            is_alive = requests.get(f"http://localhost:{free_port}/health")
        except requests.exceptions.ConnectionError:
            pass
        else:
            assert is_alive.status_code == 200
            break

        if timeout <= 0:
            raise TimeoutError("Container did not start in time")
        time.sleep(interval)
        timeout -= interval

    # Now test server (send requests and validate outputs)
    response = requests.get(f"http://localhost:{free_port}/openapi.json")
    assert response.status_code == 200

    response = requests.get(f"http://localhost:{free_port}/input_schema")
    assert response.status_code == 200
    out_input_schema = response.json()
    assert "properties" in out_input_schema

    if (
        unit_tesseract_config.volume_mounts is not None
        or unit_tesseract_config.input_path is not None
    ):
        # TODO: Mounts are not supported in HTTP mode yet, skip rest of the test for now
        return

    if unit_tesseract_config.test_with_random_inputs:
        payload_from_schema = example_from_json_schema(out_input_schema)
        response = requests.post(
            f"http://localhost:{free_port}/apply", json=payload_from_schema
        )
        assert response.status_code == 200, response.text

    sample_requests = unit_tesseract_config.sample_requests or []
    for request in sample_requests:
        if request.endpoint in ("check-gradients",):
            # Not supported in HTTP mode
            continue

        headers = {
            "Accept": f"application/{request.output_format}",
        }
        response = requests.post(
            f"http://localhost:{free_port}/{request.endpoint}",
            json=request.payload,
            headers=headers,
        )
        output = response.text
        assert response.status_code == request.expected_status_code, output

        if request.output_contains_pattern is not None:
            patterns = request.output_contains_pattern
            if isinstance(patterns, str):
                patterns = [patterns]

            for pattern in patterns:
                assert pattern in output

        if request.output_contains_array is not None:
            array = request.output_contains_array
            output_json = json.loads(output)
            assert_contains_array_allclose(output_json, array)



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import numpy as np
from pydantic import BaseModel, Field

from tesseract_core.runtime import Array, Differentiable, Float32


class InputSchema(BaseModel):
    a: Differentiable[Array[(None,), Float32]] = Field(
        description="An arbitrary vector normalized according to [...]"
    )
    b: Differentiable[Array[(None,), Float32]] = Field(
        description="An arbitrary vector. Needs to have the same dimensions as a."
    )
    s: int | float = Field(description="A scalar.", default=3)
    normalize: bool = Field(
        description="True if the output should be normalized, False otherwise.",
        default=False,
    )


class OutputSchema(BaseModel):
    result: Differentiable[Array[(None,), Float32]] = Field(
        description="Vector s·a + b"
    )


def apply(inputs: InputSchema) -> OutputSchema:
    """Multiplies a vector `a` by `s`, and sums the result to `b`."""
    a = inputs.a
    b = inputs.b
    s = inputs.s

    assert a.ndim == 1
    assert a.shape == b.shape
    result = a * s + b

    if inputs.normalize:
        norm = np.linalg.norm(result, ord=2)
        result /= norm

    return OutputSchema(result=result)


# Optional endpoints, do not remove (are used to generate complete API for docs)


def abstract_eval(abstract_inputs):
    """Calculate output shape of apply from the shape of its inputs."""
    result_shape = abstract_inputs.a
    assert result_shape is not None

    return {"result": result_shape}


def jacobian(
    inputs: InputSchema,
    jac_inputs: set[str],
    jac_outputs: set[str],
):
    assert jac_outputs == {"result"}
    n = len(inputs.a)

    partials = {}
    partials["a"] = np.eye(n) * inputs.s
    partials["b"] = np.eye(n)
    partials["s"] = inputs.a

    if inputs.normalize:
        result = inputs.a * inputs.s + inputs.b
        norm = np.linalg.norm(result, ord=2)
        partials["a"] = (
            partials["a"] / norm
            - np.outer(result, (inputs.a + inputs.s * inputs.b)) / norm**3
        )
        partials["b"] = (
            partials["b"] / norm
            - np.outer(result, (inputs.s * inputs.a + inputs.b)) / norm**3
        )
        partials["s"] = partials["s"] - (inputs.a + inputs.s * inputs.b) / norm**3 * (
            inputs.s * inputs.a * inputs.a + 2 * inputs.a * inputs.b
        )

    jacobian = {"result": {v: partials[v] for v in jac_inputs}}
    return jacobian


def jacobian_vector_product(
    inputs: InputSchema, jvp_inputs: set[str], jvp_outputs: set[str], tangent_vector
):
    return {"result": np.zeros_like(inputs.a)}


def vector_jacobian_product(
    inputs: InputSchema, vjp_inputs: set[str], vjp_outputs: set[str], cotangent_vector
):
    return {inp: np.zeros_like(inputs.a) for inp in vjp_inputs}



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import jax
from pydantic import BaseModel, Field

from tesseract_core.runtime import Differentiable, Float32, ShapeDType


def rosenbrock(x: float, y: float, a: float = 1.0, b: float = 100.0) -> float:
    return (a - x) ** 2 + b * (y - x**2) ** 2


#
# Schemas
#


class InputSchema(BaseModel):
    x: Differentiable[Float32] = Field(description="Scalar value x.", default=0.0)
    y: Differentiable[Float32] = Field(description="Scalar value y.", default=0.0)
    a: Float32 = Field(description="Scalar parameter a.", default=1.0)
    b: Float32 = Field(description="Scalar parameter b.", default=100.0)


class OutputSchema(BaseModel):
    result: Differentiable[Float32] = Field(
        description="Result of Rosenbrock function evaluation."
    )


#
# Required endpoints
#


def apply(inputs: InputSchema) -> OutputSchema:
    """Evaluates the Rosenbrock function given input values and parameters."""
    result = rosenbrock(inputs.x, inputs.y, a=inputs.a, b=inputs.b)
    return OutputSchema(result=result)


#
# Optional endpoints
#


def jacobian(
    inputs: InputSchema,
    jac_inputs: set[str],
    jac_outputs: set[str],
):
    rosenbrock_signature = ["x", "y", "a", "b"]

    jac_result = {dy: {} for dy in jac_outputs}
    for dx in jac_inputs:
        grad_func = jax.jacrev(rosenbrock, argnums=rosenbrock_signature.index(dx))
        for dy in jac_outputs:
            jac_result[dy][dx] = grad_func(inputs.x, inputs.y, inputs.a, inputs.b)

    return jac_result


def jacobian_vector_product(
    inputs: InputSchema,
    jvp_inputs: set[str],
    jvp_outputs: set[str],
    tangent_vector,
):
    # NOTE: This is a naive implementation of JVP, which is not efficient.
    jac = jacobian(inputs, jvp_inputs, jvp_outputs)
    out = {}
    for dy in jvp_outputs:
        out[dy] = sum(jac[dy][dx] * tangent_vector[dx] for dx in jvp_inputs)
    return out


def vector_jacobian_product(
    inputs: InputSchema,
    vjp_inputs: set[str],
    vjp_outputs: set[str],
    cotangent_vector,
):
    # NOTE: This is a naive implementation of VJP, which is not efficient.
    jac = jacobian(inputs, vjp_inputs, vjp_outputs)
    out = {}
    for dx in vjp_inputs:
        out[dx] = sum(jac[dy][dx] * cotangent_vector[dy] for dy in vjp_outputs)
    return out


def abstract_eval(abstract_inputs):
    """Calculate output shape of apply from the shape of its inputs."""
    return {"result": ShapeDType(shape=(), dtype="Float32")}



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import socket

import numpy as np
import pytest
from common import build_tesseract, image_exists

from tesseract_core import Tesseract
from tesseract_core.sdk import engine

expected_endpoints = {
    "apply",
    "jacobian",
    "health",
    "input_schema",
    "output_schema",
    "abstract_eval",
    "jacobian_vector_product",
    "vector_jacobian_product",
}


@pytest.fixture(scope="module")
def built_image_name(
    docker_client,
    shared_dummy_image_name,
    dummy_tesseract_location,
    docker_cleanup_module,
):
    """Build the dummy Tesseract image for the tests."""
    image_name = build_tesseract(
        docker_client, dummy_tesseract_location, shared_dummy_image_name
    )
    assert image_exists(docker_client, image_name)
    docker_cleanup_module["images"].append(image_name)
    yield image_name


def test_available_endpoints(built_image_name):
    with Tesseract.from_image(built_image_name) as vecadd:
        assert set(vecadd.available_endpoints) == expected_endpoints


def test_apply(built_image_name, dummy_tesseract_location, free_port):
    inputs = {"a": [1, 2], "b": [3, 4], "s": 1}

    # Test URL access
    tesseract_url = f"http://localhost:{free_port}"
    served_tesseract = engine.serve([built_image_name], ports=[str(free_port)])
    try:
        vecadd = Tesseract(tesseract_url)
        out = vecadd.apply(inputs)
    finally:
        engine.teardown(served_tesseract)

    assert set(out.keys()) == {"result"}
    np.testing.assert_array_equal(out["result"], np.array([4.0, 6.0]))

    # Test from_image (context manager)
    with Tesseract.from_image(built_image_name) as vecadd:
        out = vecadd.apply(inputs)

    assert set(out.keys()) == {"result"}
    np.testing.assert_array_equal(out["result"], np.array([4.0, 6.0]))

    # Test from_image (serve + teardown)
    vecadd = Tesseract.from_image(built_image_name)
    try:
        vecadd.serve()
        out = vecadd.apply(inputs)
    finally:
        vecadd.teardown()

    assert set(out.keys()) == {"result"}
    np.testing.assert_array_equal(out["result"], np.array([4.0, 6.0]))

    # Test from_tesseract_api
    with Tesseract.from_tesseract_api(
        dummy_tesseract_location / "tesseract_api.py"
    ) as vecadd:
        out = vecadd.apply(inputs)

    assert set(out.keys()) == {"result"}
    np.testing.assert_array_equal(out["result"], np.array([4.0, 6.0]))


def test_apply_with_error(built_image_name):
    # pass two inputs with different shapes, which raises an internal error
    inputs = {"a": [1, 2, 3], "b": [3, 4], "s": 1}

    with Tesseract.from_image(built_image_name) as vecadd:
        with pytest.raises(RuntimeError) as excinfo:
            vecadd.apply(inputs)

    assert "assert a.shape == b.shape" in str(excinfo.value)

    # get logs
    logs = vecadd.server_logs()
    assert "assert a.shape == b.shape" in logs


@pytest.fixture(scope="module")
def served_tesseract_remote(built_image_name):
    # Find a free port
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("", 0))
    free_port = sock.getsockname()[1]
    sock.close()
    # Serve the Tesseract image
    tesseract_url = f"http://localhost:{free_port}"
    served_tesseract = engine.serve([built_image_name], ports=[str(free_port)])
    try:
        yield tesseract_url
    finally:
        engine.teardown(served_tesseract)


@pytest.fixture(scope="module")
def served_tesseract_from_image(built_image_name):
    with Tesseract.from_image(built_image_name) as vecadd:
        yield vecadd


@pytest.fixture(scope="module")
def served_tesseract_module(dummy_tesseract_location):
    vecadd = Tesseract.from_tesseract_api(dummy_tesseract_location / "tesseract_api.py")
    yield vecadd


@pytest.mark.parametrize(
    "endpoint_name", sorted(expected_endpoints | {"openapi_schema"})
)
def test_all_endpoints(
    endpoint_name,
    served_tesseract_module,
    served_tesseract_from_image,
    served_tesseract_remote,
):
    """Test that all endpoints can be invoked without errors."""
    inputs = {"a": [1, 2], "b": [3, 4], "s": 1}

    if endpoint_name == "apply":
        inputs = {"inputs": inputs}
    elif endpoint_name == "jacobian":
        inputs = {"inputs": inputs, "jac_inputs": ["a"], "jac_outputs": ["result"]}
    elif endpoint_name == "jacobian_vector_product":
        inputs = {
            "inputs": inputs,
            "jvp_inputs": ["a"],
            "jvp_outputs": ["result"],
            "tangent_vector": {"a": np.array([1.0, 1.0])},
        }
    elif endpoint_name == "vector_jacobian_product":
        inputs = {
            "inputs": inputs,
            "vjp_inputs": ["a"],
            "vjp_outputs": ["result"],
            "cotangent_vector": {"result": np.array([1.0, 1.0])},
        }
    elif endpoint_name == "abstract_eval":
        inputs = {
            "abstract_inputs": {
                "a": {"shape": [2], "dtype": "float32"},
                "b": {"shape": [2], "dtype": "float32"},
            }
        }
    else:
        inputs = {}

    # Test from_tesseract_api
    out = getattr(served_tesseract_module, endpoint_name)
    if callable(out):
        out(**inputs)

    # Test from_image
    out = getattr(served_tesseract_from_image, endpoint_name)
    if callable(out):
        out(**inputs)

    # Test URL access
    vecadd = Tesseract(served_tesseract_remote)
    out = getattr(vecadd, endpoint_name)
    if callable(out):
        out(**inputs)



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import importlib
from collections.abc import Callable
from pathlib import Path
from types import ModuleType
from typing import Any, Union

from pydantic import BaseModel

from tesseract_core.runtime.experimental import start_run

from .config import get_config
from .schema_generation import (
    create_abstract_eval_schema,
    create_apply_schema,
    create_autodiff_schema,
)


def load_module_from_path(path: Union[Path, str]) -> ModuleType:
    """Load a module from a file path."""
    path = Path(path)

    if not path.is_file():
        raise ImportError(f"Could not load module from {path} (is not a file)")

    module_name = path.stem
    spec = importlib.util.spec_from_file_location(module_name, path)
    if spec is None:
        raise ImportError(f"Could not load module from {path}")

    module = importlib.util.module_from_spec(spec)
    try:
        spec.loader.exec_module(module)
    except Exception as exc:
        raise ImportError(f"Could not load module from {path}") from exc
    return module


def get_supported_endpoints(api_module: ModuleType) -> tuple[str, ...]:
    """Get available Tesseract functions.

    Returns:
        All optional function names defined by the Tesseract.
    """
    optional_funcs = {
        "jacobian",
        "jacobian_vector_product",
        "vector_jacobian_product",
        "abstract_eval",
    }
    return tuple(func for func in optional_funcs if hasattr(api_module, func))


def get_tesseract_api() -> ModuleType:
    """Import tesseract_api.py file."""
    return load_module_from_path(get_config().api_path)


def check_tesseract_api(api_module: ModuleType) -> None:
    """Performs basic checks on the Tesseract API module."""
    required_schemas = ("InputSchema", "OutputSchema")
    required_endpoints = ("apply",)

    for schema_name in required_schemas:
        try:
            schema = getattr(api_module, schema_name)
            if not issubclass(schema, BaseModel):
                raise TypeError(
                    f"{schema_name} is not a subclass of pydantic.BaseModel",
                )
        except AttributeError as err:
            raise TypeError(
                f"{schema_name} is not defined in Tesseract API module {api_module.__name__}",
            ) from err

    for endpoint_name in required_endpoints:
        try:
            _ = getattr(api_module, endpoint_name)
        except AttributeError as err:
            raise TypeError(
                f"{endpoint_name} is not defined in Tesseract API module {api_module.__name__}",
            ) from err


def create_endpoints(api_module: ModuleType) -> list[Callable]:
    """Create the Tesseract API endpoints.

    This ensures proper type annotations, signatures, and validation for the external-facing API.

    Args:
        api_module: The Tesseract API module.

    Returns:
        A tuple of all Tesseract API endpoints as callables.
    """
    supported_functions = get_supported_endpoints(api_module)

    endpoints = []

    def assemble_docstring(wrapped_func: Callable):
        """Decorator to assemble a docstring from multiple functions."""

        def inner(otherfunc: Callable):
            doc_parts = []
            if otherfunc.__doc__:
                doc_parts.append(otherfunc.__doc__)

            if wrapped_func.__doc__:
                doc_parts.append(wrapped_func.__doc__)

            otherfunc.__doc__ = "\n\n".join(doc_parts)
            return otherfunc

        return inner

    ApplyInputSchema, ApplyOutputSchema = create_apply_schema(
        api_module.InputSchema, api_module.OutputSchema
    )

    @assemble_docstring(api_module.apply)
    def apply(payload: ApplyInputSchema) -> ApplyOutputSchema:
        """Apply the Tesseract to the input data."""
        with start_run():
            out = api_module.apply(payload.inputs)
        if isinstance(out, api_module.OutputSchema):
            out = out.model_dump()
        return ApplyOutputSchema.model_validate(out)

    endpoints.append(apply)

    if "jacobian" in supported_functions:
        JacobianInputSchema, JacobianOutputSchema = create_autodiff_schema(
            api_module.InputSchema, api_module.OutputSchema, ad_flavor="jacobian"
        )

        @assemble_docstring(api_module.jacobian)
        def jacobian(payload: JacobianInputSchema) -> JacobianOutputSchema:
            """Computes the Jacobian of the Tesseract.

            Differentiates ``jac_outputs`` with respect to ``jac_inputs``, at the point ``inputs``.
            """
            with start_run():
                out = api_module.jacobian(**dict(payload))
            return JacobianOutputSchema.model_validate(
                out,
                context={
                    "output_keys": payload.jac_outputs,
                    "input_keys": payload.jac_inputs,
                },
            )

        endpoints.append(jacobian)

    if "jacobian_vector_product" in supported_functions:
        JVPInputSchema, JVPOutputSchema = create_autodiff_schema(
            api_module.InputSchema, api_module.OutputSchema, ad_flavor="jvp"
        )

        @assemble_docstring(api_module.jacobian_vector_product)
        def jacobian_vector_product(payload: JVPInputSchema) -> JVPOutputSchema:
            """Compute the Jacobian vector product of the Tesseract at the input data.

            Evaluates the Jacobian vector product between the Jacobian given by ``jvp_outputs``
            with respect to ``jvp_inputs`` at the point ``inputs`` and the given tangent vector.
            """
            with start_run():
                out = api_module.jacobian_vector_product(**dict(payload))
            return JVPOutputSchema.model_validate(
                out, context={"output_keys": payload.jvp_outputs}
            )

        endpoints.append(jacobian_vector_product)

    if "vector_jacobian_product" in supported_functions:
        VJPInputSchema, VJPOutputSchema = create_autodiff_schema(
            api_module.InputSchema, api_module.OutputSchema, ad_flavor="vjp"
        )

        @assemble_docstring(api_module.vector_jacobian_product)
        def vector_jacobian_product(payload: VJPInputSchema) -> VJPOutputSchema:
            """Compute the Jacobian vector product of the Tesseract at the input data.

            Computes the vector Jacobian product between the Jacobian given by ``vjp_outputs``
            with respect to ``vjp_inputs`` at the point ``inputs`` and the given cotangent vector.
            """
            with start_run():
                out = api_module.vector_jacobian_product(**dict(payload))
            return VJPOutputSchema.model_validate(
                out, context={"input_keys": payload.vjp_inputs}
            )

        endpoints.append(vector_jacobian_product)

    def health() -> dict[str, Any]:
        """Get health status of the Tesseract instance."""
        return {"status": "ok"}

    endpoints.append(health)

    def input_schema() -> dict[str, Any]:
        """Get input schema for tesseract apply function."""
        return ApplyInputSchema.model_json_schema()

    endpoints.append(input_schema)

    def output_schema() -> dict[str, Any]:
        """Get output schema for tesseract apply function."""
        return ApplyOutputSchema.model_json_schema()

    endpoints.append(output_schema)

    if "abstract_eval" in supported_functions:
        AbstractEvalInputSchema, AbstractEvalOutputSchema = create_abstract_eval_schema(
            api_module.InputSchema, api_module.OutputSchema
        )

        @assemble_docstring(api_module.abstract_eval)
        def abstract_eval(payload: AbstractEvalInputSchema) -> AbstractEvalOutputSchema:
            """Perform abstract evaluation of the Tesseract on the input data."""
            out = api_module.abstract_eval(payload.inputs)
            return AbstractEvalOutputSchema.model_validate(out)

        endpoints.append(abstract_eval)

    return endpoints
